import pytest

from tests.sql_models_data import USE_SQLA

pytestmark = pytest.mark.skipif(not USE_SQLA, reason='For SqlAlchemy only')

if USE_SQLA:
    from dropland.storages.containers import SqlStorage, SingleSqlStorage, MultipleSqlStorage, default_sql_storage
    from dropland.storages.sql import SqlStorageType, EngineConfig, SqlStorageEngine
    from tests import MYSQL_URI, POSTGRES_URI, SQLITE_URI


@pytest.mark.asyncio
async def test_create_engine():
    sqlite_config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )
    pg_config = EngineConfig(
        url=POSTGRES_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )
    mysql_config = EngineConfig(
        url=MYSQL_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine_factory = default_sql_storage.engine_factory()

    assert engine_factory.create_engine(SqlStorageType.SQLITE, sqlite_config, use_async=False)
    assert engine_factory.create_engine(SqlStorageType.SQLITE, sqlite_config, use_async=True)
    assert engine_factory.create_engine(SqlStorageType.POSTGRES, pg_config, use_async=False)
    assert engine_factory.create_engine(SqlStorageType.POSTGRES, pg_config, use_async=True)
    assert engine_factory.create_engine(SqlStorageType.MYSQL, mysql_config, use_async=False)
    assert engine_factory.create_engine(SqlStorageType.MYSQL, mysql_config, use_async=True)

    for db_type in (SqlStorageType.SQLITE, SqlStorageType.POSTGRES, SqlStorageType.MYSQL):
        for e in engine_factory.get_engines_for_type(db_type, is_async=False):
            assert e.db_type == db_type
            assert e.engine
            assert not e.engine.is_async

        for e in engine_factory.get_engines_for_type(db_type, is_async=True):
            assert e.db_type == db_type
            assert e.engine
            assert e.engine.is_async

    for e in engine_factory.get_engines_for_backend('sqla', is_async=False):
        assert e.backend == 'sqla'
        assert e.engine

    for e in engine_factory.get_engines_for_backend('sqla', is_async=True):
        assert e.backend == 'sqla'
        assert e.engine

    assert engine_factory.get_backend_names() == ['sqla']


@pytest.mark.asyncio
async def test_create_sqlite_engine():
    config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine = default_sql_storage.create_engine(SqlStorageType.SQLITE, config, use_async=False)
    assert engine
    engine.start()

    with engine.new_connection() as conn:
        res = conn.execute('select sqlite_version();')
        print(res.fetchone()[0])
        res = conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    engine.stop()

    engine = default_sql_storage.create_engine(SqlStorageType.SQLITE, config, use_async=True)
    assert engine
    await engine.async_start()

    async with engine.new_connection() as conn:
        res = await conn.execute('select sqlite_version();')
        print(res.fetchone()[0])
        res = await conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    await engine.async_stop()


@pytest.mark.asyncio
async def test_create_pg_engine():
    config = EngineConfig(
        url=POSTGRES_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine = default_sql_storage.create_engine(SqlStorageType.POSTGRES, config, use_async=False)
    assert engine
    engine.start()

    with engine.new_connection() as conn:
        res = conn.execute('select version();')
        print(res.fetchone()[0])
        res = conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    engine.stop()

    engine = default_sql_storage.create_engine(SqlStorageType.POSTGRES, config, use_async=True)
    assert engine
    await engine.async_start()

    async with engine.new_connection() as conn:
        res = await conn.execute('select version();')
        print(res.fetchone()[0])
        res = await conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    await engine.async_stop()


@pytest.mark.asyncio
async def test_create_mysql_engine():
    config = EngineConfig(
        url=MYSQL_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine = default_sql_storage.create_engine(SqlStorageType.MYSQL, config, use_async=False)
    assert engine
    engine.start()

    with engine.new_connection() as conn:
        res = conn.execute('select version();')
        print(res.fetchone()[0])
        res = conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    engine.stop()

    engine = default_sql_storage.create_engine(SqlStorageType.MYSQL, config, use_async=True)
    assert engine
    await engine.async_start()

    async with engine.new_connection() as conn:
        res = await conn.execute('select version();')
        print(res.fetchone()[0])
        res = await conn.execute('select 1 + 2;')
        assert res.fetchone()[0] == 3

    await engine.async_stop()


@pytest.mark.asyncio
async def test_storage_container():
    sqlite_config = EngineConfig(
        url=SQLITE_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    cont = SqlStorage()

    eng = cont.create_engine(SqlStorageType.SQLITE, sqlite_config, use_async=False)
    assert isinstance(eng, SqlStorageEngine)
    cont.unwire()

    cont = SingleSqlStorage()
    cont.config.from_dict({
        'db_type': SqlStorageType.SQLITE,
        'engine_config': sqlite_config,
        'use_async': False
    })
    eng1 = cont.create_engine()
    eng2 = cont.create_engine()
    assert eng1 is eng2
    cont.unwire()

    cont = MultipleSqlStorage()
    cont.config.from_dict({
        'one': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': False
        },
        'two': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': sqlite_config,
            'use_async': True
        },
    })

    eng1 = cont.create_engine('one')
    eng2 = cont.create_engine('two')
    assert eng1 is not eng2
    assert eng1.is_async is False
    assert eng2.is_async is True

    cont.unwire()



from typing import List

from fastapi import FastAPI, Request, Response

from dropland.storages.containers import with_async_engines


def add_middleware(
        app: FastAPI, sql_backends: List[str] = None, redis_names: List[str] = None,
        begin_sql_tx: bool = True, autocommit: bool = True):
    @app.middleware('http')
    async def __set_data_source_middleware(request: Request, call_next) -> Response:
        async with with_async_engines(sql_backends, redis_names, begin_sql_tx, autocommit):
            return await call_next(request)



SQLITE_URI = 'sqlite:///db.sqlite'
POSTGRES_URI = 'postgresql://postgres:postgres@localhost:5432/postgres'
MYSQL_URI = 'mysql://mysql:mysql@localhost:3306/mysql'
REDIS_URI = 'redis://localhost:6379/0'



import asyncio

import pytest

from dropland.storages.redis import USE_REDIS
from dropland.storages.sql import USE_SQL

if USE_REDIS:
    from dropland.storages.containers import default_redis_storage
    from dropland.storages.redis import EngineConfig as RedisEngineConfig

if USE_SQL:
    from dropland.storages.containers import MultipleSqlStorage
    from dropland.storages.sql import EngineConfig as SqlEngineConfig, SqlStorageType

from tests import MYSQL_URI, POSTGRES_URI, REDIS_URI, SQLITE_URI


if USE_SQL:
    sql_storage = MultipleSqlStorage()

    sql_storage.config.from_dict({
        'sqlite': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': SqlEngineConfig(
                url=SQLITE_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': False
        },
        'asqlite': {
            'db_type': SqlStorageType.SQLITE,
            'engine_config': SqlEngineConfig(
                url=SQLITE_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True
        },
        'pg': {
            'db_type': SqlStorageType.POSTGRES,
            'engine_config': SqlEngineConfig(
                url=POSTGRES_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': False
        },
        'apg': {
            'db_type': SqlStorageType.POSTGRES,
            'engine_config': SqlEngineConfig(
                url=POSTGRES_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True
        },
        'apg-gino': {
            'db_type': SqlStorageType.POSTGRES,
            'engine_config': SqlEngineConfig(
                url=POSTGRES_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True,
            'backend_name': 'gino'
        },
        'ms': {
            'db_type': SqlStorageType.MYSQL,
            'engine_config': SqlEngineConfig(
                url=MYSQL_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': False
        },
        'ams': {
            'db_type': SqlStorageType.MYSQL,
            'engine_config': SqlEngineConfig(
                url=MYSQL_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True
        },
        'ams-gino': {
            'db_type': SqlStorageType.MYSQL,
            'engine_config': SqlEngineConfig(
                url=MYSQL_URI,
                pool_min_size=1, pool_max_size=2,
                pool_expire_seconds=15, pool_timeout_seconds=15,
            ),
            'use_async': True,
            'backend_name': 'gino'
        },
    })


@pytest.fixture(scope='session')
def test_sql_storage():
    return sql_storage


@pytest.fixture(scope='session')
def test_redis_storage():
    if USE_REDIS:
        return default_redis_storage
    else:
        return None


@pytest.fixture(scope='session')
def sqlite_engine(test_sql_storage):
    return test_sql_storage.create_engine('sqlite')


@pytest.fixture(scope='session')
def sqlite_async_engine(test_sql_storage):
    return test_sql_storage.create_engine('asqlite')


@pytest.fixture(scope='session')
def pg_engine(test_sql_storage):
    return test_sql_storage.create_engine('pg')


@pytest.fixture(scope='session')
def pg_async_engine(test_sql_storage):
    return test_sql_storage.create_engine('apg')


@pytest.fixture(scope='session')
def mysql_engine(test_sql_storage):
    return test_sql_storage.create_engine('ms')


@pytest.fixture(scope='session')
def mysql_async_engine(test_sql_storage):
    return test_sql_storage.create_engine('ams')


@pytest.fixture(scope='session')
def pg_gino_engine(test_sql_storage):
    return test_sql_storage.create_engine('apg-gino')


@pytest.fixture(scope='session')
def mysql_gino_engine(test_sql_storage):
    return test_sql_storage.create_engine('ams-gino')


@pytest.fixture(scope='session')
def redis_engine(test_redis_storage):
    engine_config = RedisEngineConfig(url=REDIS_URI)
    return test_redis_storage.create_engine('dropland', engine_config)


@pytest.fixture(scope='session')
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


if USE_SQL:
    @pytest.fixture(scope='function')
    async def _sql_ctx(event_loop, test_sql_storage):
        if USE_SQL:
            async with test_sql_storage.manager().with_async_engines(autocommit=False) as ctx:
                yield ctx
        else:
            yield


    @pytest.fixture(scope='function')
    def sql_ctx(_sql_ctx, test_sql_storage):
        with test_sql_storage.manager().set_context(_sql_ctx):
            yield _sql_ctx


if USE_REDIS:
    @pytest.fixture(scope='function')
    async def _redis_ctx(event_loop, test_redis_storage):
        if USE_REDIS:
            async with test_redis_storage.manager().with_async_engines() as ctx:
                yield ctx
        else:
            yield


    @pytest.fixture(scope='function')
    def redis_ctx(_redis_ctx, test_redis_storage):
        with test_redis_storage.manager().set_context(_redis_ctx):
            yield _redis_ctx


if USE_SQL and not USE_REDIS:
    @pytest.fixture(scope='function')
    async def data_ctx(sql_ctx):
        yield sql_ctx, None

elif not USE_SQL and USE_REDIS:
    @pytest.fixture(scope='function')
    async def data_ctx(redis_ctx):
        yield None, redis_ctx

elif USE_SQL and USE_REDIS:
    @pytest.fixture(scope='function')
    async def data_ctx(sql_ctx, redis_ctx):
        yield sql_ctx, redis_ctx

else:
    @pytest.fixture(scope='function')
    async def data_ctx():
        yield None, None



import contextlib
from datetime import timedelta
from typing import List

from dependency_injector import containers, providers
from dependency_injector.wiring import Provide, inject
from pytimeparse.timeparse import timeparse

from dropland.storages.redis import USE_REDIS
from dropland.storages.sql import USE_SQL
from dropland.util import default_value

if USE_REDIS:
    from dropland.storages.redis.connections import ConnectionManager as RedisConnectionManager
    from dropland.storages.redis.engine import RedisStorageBackend, EngineConfig as RedisEngineConfig

if USE_SQL:
    from dropland.storages.sql.connections import ConnectionManager as SqlConnectionManager
    from dropland.storages.sql.engine import EngineFactory as SqlEngineFactory, EngineConfig as SqlEngineConfig
    from dropland.storages.sql.loader import SqlBackendLoader


if USE_SQL:
    import dropland.storages.sql.model

    class SqlStorage(containers.DeclarativeContainer):
        __self__ = providers.Self()
        backend_loader = providers.Singleton(SqlBackendLoader)
        engine_factory = providers.Singleton(SqlEngineFactory, backend_loader)
        manager = providers.Singleton(SqlConnectionManager, engine_factory)

        def _create_engine(self, *args, **kwargs):
            return self.engine_factory().create_engine(*args, **kwargs)

        create_engine = providers.Factory(_create_engine, __self__)
        wiring_config = containers.WiringConfiguration(modules=[dropland.storages.sql.model, __name__])


    class SingleSqlStorage(SqlStorage):
        __self__ = providers.Self()
        config = providers.Configuration()

        def _create_engine(self):
            if isinstance(self.config.engine_config(), SqlEngineConfig):
                engine_config = self.config.engine_config()
            else:
                engine_config = SqlEngineConfig(
                    url=self.config.engine_config.url(),
                    echo=self.config.engine_config.echo.as_(bool)(),
                    pool_min_size=self.config.engine_config.
                        pool_min_size.as_(default_value(int))(default=1),
                    pool_max_size=self.config.engine_config.
                        pool_max_size.as_(default_value(int))(default=8),
                    pool_expire_seconds=self.config.engine_config.
                        pool_expire_seconds.as_(default_value(int))(default=60),
                    pool_timeout_seconds=self.config.engine_config.
                        pool_timeout_seconds.as_(default_value(int))(default=15)
                )
            return SqlStorage._create_engine(
                self, self.config.db_type(), engine_config, backend_name=self.config.backend_name(),
                use_async=self.config.use_async.as_(bool)())

        create_engine = providers.Factory(_create_engine, __self__)
        wiring_config = containers.WiringConfiguration(modules=[dropland.storages.sql.model, __name__])


    class MultipleSqlStorage(SqlStorage):
        __self__ = providers.Self()
        config = providers.Configuration()

        def _create_engine(self, name: str):
            if conf := self.config.get(name):
                if isinstance(conf['engine_config'], SqlEngineConfig):
                    engine_config = conf['engine_config']
                else:
                    engine_config = SqlEngineConfig(
                        url=conf['engine_config']['url'],
                        echo=bool(conf['engine_config'].get('echo')),
                        pool_min_size=int(conf['engine_config'].get('pool_min_size', 1)),
                        pool_max_size=int(conf['engine_config'].get('pool_max_size', 8)),
                        pool_expire_seconds=int(conf['engine_config'].get('pool_expire_seconds', 60)),
                        pool_timeout_seconds=int(conf['engine_config'].get('pool_timeout_seconds', 15))
                    )
                return SqlStorage._create_engine(
                    self, conf['db_type'], engine_config, backend_name=conf.get('backend_name'),
                    use_async=conf.get('use_async', False))
            return None

        create_engine = providers.Factory(_create_engine, __self__)
        wiring_config = containers.WiringConfiguration(modules=[dropland.storages.sql.model, __name__])

    default_sql_storage = SqlStorage()
else:
    class SqlStorage:
        manager = None


if USE_REDIS:
    import dropland.storages.redis.model

    class RedisStorage(containers.DeclarativeContainer):
        __self__ = providers.Self()
        engine_factory = providers.Singleton(RedisStorageBackend)
        manager = providers.Singleton(RedisConnectionManager, engine_factory)
        default_ttl = providers.Object(timedelta(seconds=timeparse('1 min')))

        def _create_engine(self, *args, **kwargs):
            return self.engine_factory().create_engine(*args, **kwargs)

        create_engine = providers.Factory(_create_engine, __self__)
        wiring_config = containers.WiringConfiguration(modules=[dropland.storages.redis.model, __name__])


    class SingleRedisStorage(RedisStorage):
        __self__ = providers.Self()
        config = providers.Configuration()
        default_ttl = providers.Object(timedelta(
            seconds=timeparse(config.get('default_cache_ttl', required=False) or '1 min')))

        def _create_engine(self):
            if isinstance(self.config.engine_config(), RedisEngineConfig):
                engine_config = self.config.engine_config()
            else:
                engine_config = RedisEngineConfig(
                    url=self.config.engine_config.url(),
                    max_connections=self.config.engine_config.
                        max_connections.as_(default_value(int))(default=4),
                    pool_timeout_seconds=self.config.engine_config.
                        pool_timeout_seconds.as_(default_value(int))(default=5)
                )
            return RedisStorage._create_engine(self, self.config.name(), engine_config, self.default_ttl())

        create_engine = providers.Factory(_create_engine, __self__)
        wiring_config = containers.WiringConfiguration(modules=[dropland.storages.redis.model, __name__])


    class MultipleRedisStorage(RedisStorage):
        __self__ = providers.Self()
        config = providers.Configuration()

        def _create_engine(self, name: str):
            if conf := self.config.get(name):
                if isinstance(conf['engine_config'], RedisEngineConfig):
                    engine_config = conf['engine_config']
                else:
                    engine_config = RedisEngineConfig(
                        url=conf['engine_config']['url'],
                        max_connections=int(conf['engine_config'].get('max_connections', 4)),
                        pool_timeout_seconds=int(conf['engine_config'].get('pool_timeout_seconds', 5))
                    )
                return RedisStorage._create_engine(
                    self, conf['name'], engine_config,
                    timedelta(seconds=timeparse(conf.get('default_cache_ttl', '1 min'))))
            return None

        create_engine = providers.Factory(_create_engine, __self__)
        wiring_config = containers.WiringConfiguration(modules=[dropland.storages.redis.model, __name__])

    default_redis_storage = RedisStorage()
else:
    class RedisStorage:
        manager = None


if USE_SQL or USE_REDIS:
    @inject
    def use_engines(
            sql_backends: List[str] = None,
            redis_names: List[str] = None,
            begin_sql_tx: bool = True,
            autocommit: bool = True,
            sql_manager: 'SqlConnectionManager' = Provide[SqlStorage.manager] if USE_SQL else None,
            redis_manager: 'RedisConnectionManager' = Provide[RedisStorage.manager] if USE_REDIS else None
            ):
        if USE_REDIS and isinstance(redis_manager, RedisConnectionManager):
            redis_inner = redis_manager.use_engines(redis_names)
        if USE_SQL and isinstance(sql_manager, SqlConnectionManager):
            sql_inner = sql_manager.use_engines(sql_backends, begin_sql_tx, autocommit)

        def inner(function):
            if USE_REDIS and isinstance(redis_manager, RedisConnectionManager):
                function = redis_inner(function)
            if USE_SQL and isinstance(sql_manager, SqlConnectionManager):
                function = sql_inner(function)
            return function

        return inner

    @contextlib.asynccontextmanager
    @inject
    async def with_async_engines(
            sql_backends: List[str] = None,
            redis_names: List[str] = None,
            begin_sql_tx: bool = True,
            autocommit: bool = True,
            sql_manager: 'SqlConnectionManager' = Provide[SqlStorage.manager] if USE_SQL else None,
            redis_manager: 'RedisConnectionManager' = Provide[RedisStorage.manager] if USE_REDIS else None
            ):
        async with contextlib.AsyncExitStack() as stack:
            if USE_REDIS and isinstance(redis_manager, RedisConnectionManager):
                await stack.enter_async_context(redis_manager.with_async_engines(redis_names))
            if USE_SQL and isinstance(sql_manager, SqlConnectionManager):
                await stack.enter_async_context(sql_manager.with_async_engines(sql_backends, begin_sql_tx, autocommit))
            yield



import os
from dataclasses import dataclass
from datetime import timedelta
from typing import Dict, List, Optional, Union

import pytz
from apscheduler.schedulers.base import BaseScheduler

from dropland.log import logger, tr
from dropland.storages.redis import USE_REDIS
from dropland.storages.sql import SqlStorageEngine


@dataclass
class EngineConfig:
    sql_url: Union[str, SqlStorageEngine] = None
    sql_tablename: Optional[str] = None
    redis_url: Optional[str] = None
    redis_job_key: Optional[str] = None

    job_coalesce: bool = False
    job_max_instances: int = 1
    job_misfire_grace_time: int = 24 * 3600

    task_host: str = '0.0.0.0'
    task_port: int = 3000
    task_processes: int = os.cpu_count()
    task_workers: int = os.cpu_count()
    task_rpc_timeout_seconds: int = 5
    create_remote_engine: Optional[bool] = None
    timezone: str = 'UTC'


class TaskManagerBackend:
    def __init__(self, is_scheduler: bool):
        self._engines: Dict[str, BaseScheduler] = dict()
        self._is_scheduler = is_scheduler

    @property
    def name(self) -> str:
        return 'scheduler'

    def create_engine(self, name: str, config: EngineConfig) -> Optional[BaseScheduler]:
        if engine := self._engines.get(name):
            return engine

        from apscheduler.jobstores.memory import MemoryJobStore

        jobstores = {'default': MemoryJobStore()}

        if USE_REDIS and config.redis_url:
            from apscheduler.jobstores.redis import RedisJobStore
            from redis import ConnectionPool

            jobstores['redis'] = RedisJobStore(
                connection_pool=ConnectionPool.from_url(config.redis_url),
                jobs_key=f'{config.redis_job_key}.aps.jobs'
                    if isinstance(config.redis_job_key, str) else 'apscheduler.jobs',
                run_times_key=f'{config.redis_job_key}.aps.run_times'
                    if isinstance(config.redis_job_key, str) else 'apscheduler.run_times',
            )

        if config.sql_url:
            from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore

            jobstores['sql'] = SQLAlchemyJobStore(
                url=config.sql_url if isinstance(config.sql_url, str) else None,
                engine=config.sql_url if not isinstance(config.sql_url, str) else None,
                metadata=config.sql_url.metadata if not isinstance(config.sql_url, str) else None,
                tablename=config.sql_tablename if config.sql_tablename else 'apscheduler_jobs',
                engine_options={'pool_pre_ping': True},
            )

        job_defaults = {
            'coalesce': config.job_coalesce,
            'max_instances': config.job_max_instances,
            'misfire_grace_time': config.job_misfire_grace_time,
        }

        engine = self._create_engine(config, jobstores=jobstores, job_defaults=job_defaults)
        self._engines[name] = engine
        logger.info(tr('dropland.scheduler.created'))
        return engine

    def get_engine(self, name: str) -> Optional[BaseScheduler]:
        return self._engines.get(name)

    def get_engine_names(self) -> List[str]:
        return list(self._engines.keys())

    def _create_engine(self, config: EngineConfig, **kwargs):
        from apscheduler.executors.asyncio import AsyncIOExecutor
        from apscheduler.executors.pool import ProcessPoolExecutor
        from concurrent.futures.thread import ThreadPoolExecutor as BackgroundExecutor
        from dropland.tasks.local import Scheduler

        logger.info(tr('dropland.scheduler.create.local'))

        executors = {
            'default': AsyncIOExecutor(),
            'process': ProcessPoolExecutor(config.task_processes),
        }

        if config.timezone:
            timezone = pytz.timezone(config.timezone) if isinstance(config.timezone, str) else config.timezone
        else:
            timezone = None

        create_remote_engine = config.create_remote_engine and not self._is_scheduler
        executor = BackgroundExecutor(config.task_workers, thread_name_prefix='TaskWorker')
        scheduler = Scheduler(
            remote=self._create_remote(config) if create_remote_engine else None,
            executor=executor, executors=executors, timezone=timezone, **kwargs
        )
        return scheduler

    def _create_remote(self, config: EngineConfig):
        import apscheduler.jobstores.base
        from dropland.tasks.remote import RemoteScheduler, register_exception

        logger.info(tr('dropland.scheduler.create.remote'))

        register_exception(apscheduler.jobstores.base.JobLookupError)
        register_exception(apscheduler.jobstores.base.ConflictingIdError)
        register_exception(apscheduler.jobstores.base.TransientJobError)

        return RemoteScheduler(
            host=config.task_host, port=config.task_port,
            rpc_timeout=timedelta(seconds=config.task_rpc_timeout_seconds)
        )



from dropland.tr.static import StaticTranslator


def test_load_file_unload_file():
    tr = StaticTranslator()
    assert tr.size() == 0

    tr.load_file('tests/translations.toml')
    assert tr.size() == 6
    tr.load_file('tests/translations.toml', namespace='root')
    assert tr.size() == 12

    assert 'Title' == tr('translations.title')
    assert 'Title' == tr('root.title')

    assert not tr.set_default_lang('translations1', 'ru')
    assert tr.set_default_lang('translations', 'ru')

    assert 'Заголовок' == tr('translations.title')
    assert 'Title' == tr('root.title')

    tr.unload('translations1')
    tr.unload('translations')

    assert tr.size() == 6
    assert not tr('translations.title')
    assert 'Title' == tr('root.title')


def test_one_namespace():
    tr = StaticTranslator()
    tr.load_file('tests/translations.toml')

    assert not tr('default')
    assert 'Title' == tr('title')
    assert 'Title' == tr('translations.title')
    assert 'Title' == tr('title', 'en')
    assert 'Заголовок' == tr('title', 'ru')
    assert 'Title' == tr('translations.title')
    assert 'Title' == tr('translations.title', 'en')
    assert 'Заголовок' == tr('translations.title', 'ru')

    assert 'Name' == tr('owner.name')
    assert 'Name' == tr('owner.name', 'en')
    assert 'Имя' == tr('owner.name', 'ru')
    assert 'Name' == tr('translations.owner.name')
    assert 'Name' == tr('translations.owner.name', 'en')
    assert 'Имя' == tr('translations.owner.name', 'ru')

    assert 'Name' == tr('owner.name.en')
    assert 'Имя' == tr('owner.name.ru')
    assert 'Name' == tr('translations.owner.name.en')
    assert 'Имя' == tr('translations.owner.name.ru')

    assert 'Name' == tr('owner.name.en', 'en')
    assert 'Имя' == tr('owner.name.ru', 'en')
    assert 'Name' == tr('owner.name.en', 'ru')
    assert 'Имя' == tr('owner.name.ru', 'ru')
    assert 'Name' == tr('translations.owner.name.en', 'en')
    assert 'Имя' == tr('translations.owner.name.ru', 'en')
    assert 'Name' == tr('translations.owner.name.en', 'ru')
    assert 'Имя' == tr('translations.owner.name.ru', 'ru')

    assert 'Data' == tr('data.title1')
    assert 'Data' == tr('data.title1', 'en')
    assert 'Data' == tr('data.title1', 'ru')
    assert 'Data' == tr('translations.data.title1')
    assert 'Data' == tr('translations.data.title1', 'en')
    assert 'Data' == tr('translations.data.title1', 'ru')

    assert not tr('data.title1.en')
    assert not tr('data.title1.ru')
    assert not tr('data.title1.en', 'en')
    assert not tr('data.title1.en', 'ru')
    assert not tr('data.title1.ru', 'en')
    assert not tr('data.title1.ru', 'ru')
    assert not tr('translations.data.title1.en')
    assert not tr('translations.data.title1.ru')
    assert not tr('translations.data.title1.en', 'en')
    assert not tr('translations.data.title1.en', 'ru')
    assert not tr('translations.data.title1.ru', 'en')
    assert not tr('translations.data.title1.ru', 'ru')

    assert not tr('data.title2')
    assert not tr('data.title2', 'en')
    assert 'Данные' == tr('data.title2', 'ru')
    assert not tr('translations.data.title2')
    assert not tr('translations.data.title2', 'en')
    assert 'Данные' == tr('translations.data.title2', 'ru')

    assert not tr('data.title2.en')
    assert 'Данные' == tr('data.title2.ru')
    assert not tr('data.title2.en', 'en')
    assert not tr('data.title2.en', 'ru')
    assert 'Данные' == tr('data.title2.ru', 'en')
    assert 'Данные' == tr('data.title2.ru', 'ru')
    assert not tr('translations.data.title2.en')
    assert 'Данные' == tr('translations.data.title2.ru')
    assert not tr('translations.data.title2.en', 'en')
    assert not tr('translations.data.title2.en', 'ru')
    assert 'Данные' == tr('translations.data.title2.ru', 'en')
    assert 'Данные' == tr('translations.data.title2.ru', 'ru')


def test_two_namespaces():
    tr = StaticTranslator()
    tr.load_file('tests/translations.toml')
    tr.load_file('tests/translations.toml', namespace='root')

    assert not tr('default')
    assert not tr('root.default')
    assert not tr('title')
    assert 'Title' == tr('translations.title')
    assert 'Title' == tr('root.title')
    assert not tr('title', 'en')
    assert not tr('title', 'ru')
    assert 'Title' == tr('translations.title', 'en')
    assert 'Заголовок' == tr('translations.title', 'ru')
    assert 'Title' == tr('root.title', 'en')
    assert 'Заголовок' == tr('root.title', 'ru')

    assert not tr('owner.name')
    assert not tr('owner.name', 'en')
    assert not tr('owner.name', 'ru')
    assert 'Name' == tr('translations.owner.name')
    assert 'Name' == tr('translations.owner.name', 'en')
    assert 'Имя' == tr('translations.owner.name', 'ru')
    assert 'Name' == tr('root.owner.name')
    assert 'Name' == tr('root.owner.name', 'en')
    assert 'Имя' == tr('root.owner.name', 'ru')

    assert not tr('owner.name.en')
    assert not tr('owner.name.ru')
    assert 'Name' == tr('translations.owner.name.en')
    assert 'Имя' == tr('translations.owner.name.ru')
    assert 'Name' == tr('root.owner.name.en')
    assert 'Имя' == tr('root.owner.name.ru')

    assert not tr('owner.name.en', 'en')
    assert not tr('owner.name.ru', 'en')
    assert not tr('owner.name.en', 'ru')
    assert not tr('owner.name.ru', 'ru')
    assert 'Name' == tr('translations.owner.name.en', 'en')
    assert 'Имя' == tr('translations.owner.name.ru', 'en')
    assert 'Name' == tr('translations.owner.name.en', 'ru')
    assert 'Имя' == tr('translations.owner.name.ru', 'ru')
    assert 'Name' == tr('root.owner.name.en', 'en')
    assert 'Имя' == tr('root.owner.name.ru', 'en')
    assert 'Name' == tr('root.owner.name.en', 'ru')
    assert 'Имя' == tr('root.owner.name.ru', 'ru')

    assert not tr('data.title1')
    assert not tr('data.title1', 'en')
    assert not tr('data.title1', 'ru')
    assert 'Data' == tr('translations.data.title1')
    assert 'Data' == tr('translations.data.title1', 'en')
    assert 'Data' == tr('translations.data.title1', 'ru')
    assert 'Data' == tr('root.data.title1')
    assert 'Data' == tr('root.data.title1', 'en')
    assert 'Data' == tr('root.data.title1', 'ru')

    assert not tr('data.title1.en')
    assert not tr('data.title1.ru')
    assert not tr('data.title1.en', 'en')
    assert not tr('data.title1.en', 'ru')
    assert not tr('data.title1.ru', 'en')
    assert not tr('data.title1.ru', 'ru')
    assert not tr('translations.data.title1.en')
    assert not tr('translations.data.title1.ru')
    assert not tr('translations.data.title1.en', 'en')
    assert not tr('translations.data.title1.en', 'ru')
    assert not tr('translations.data.title1.ru', 'en')
    assert not tr('translations.data.title1.ru', 'ru')
    assert not tr('root.data.title1.en')
    assert not tr('root.data.title1.ru')
    assert not tr('root.data.title1.en', 'en')
    assert not tr('root.data.title1.en', 'ru')
    assert not tr('root.data.title1.ru', 'en')
    assert not tr('root.data.title1.ru', 'ru')

    assert not tr('data.title2')
    assert not tr('data.title2', 'en')
    assert not tr('data.title2', 'ru')
    assert not tr('translations.data.title2')
    assert not tr('translations.data.title2', 'en')
    assert 'Данные' == tr('translations.data.title2', 'ru')
    assert not tr('root.data.title2')
    assert not tr('root.data.title2', 'en')
    assert 'Данные' == tr('root.data.title2', 'ru')

    assert not tr('data.title2.en')
    assert not tr('data.title2.ru')
    assert not tr('data.title2.en', 'en')
    assert not tr('data.title2.en', 'ru')
    assert not tr('data.title2.ru', 'en')
    assert not tr('data.title2.ru', 'ru')
    assert not tr('translations.data.title2.en')
    assert 'Данные' == tr('translations.data.title2.ru')
    assert not tr('translations.data.title2.en', 'en')
    assert not tr('translations.data.title2.en', 'ru')
    assert 'Данные' == tr('translations.data.title2.ru', 'en')
    assert 'Данные' == tr('translations.data.title2.ru', 'ru')
    assert not tr('root.data.title2.en')
    assert 'Данные' == tr('root.data.title2.ru')
    assert not tr('root.data.title2.en', 'en')
    assert not tr('root.data.title2.en', 'ru')
    assert 'Данные' == tr('root.data.title2.ru', 'en')
    assert 'Данные' == tr('root.data.title2.ru', 'ru')



from dropland.storages.redis import USE_REDIS

if USE_REDIS:
    from dropland.storages.containers import default_redis_storage
    from dropland.storages.redis import EngineConfig

from tests import REDIS_URI

if USE_REDIS:
    redis_engine = default_redis_storage.create_engine('dropland', EngineConfig(url=REDIS_URI))



import asyncio
import contextlib
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Tuple

from contextvars import ContextVar

from dropland.storages.sql.base import SqlStorageEngine, SqlStorageType
from dropland.storages.sql.engine import EngineKey, EngineFactory


@dataclass
class ConnectionData:
    engine: SqlStorageEngine
    connection: Any
    timeout_secs: int
    in_transaction: bool = False


class ConnectionContext:
    def __init__(self):
        self.conns: Dict[EngineKey, ConnectionData] = dict()
        self.conns_by_type: Dict[SqlStorageType, Set[str]] = defaultdict(set)
        self.conns_by_backend: Dict[str, Set[SqlStorageType]] = defaultdict(set)


class ConnectionManager:
    def __init__(self, engine_factory: EngineFactory):
        self._ctx: ContextVar[ConnectionContext] = ContextVar('_ctx')
        self._engine_factory = engine_factory

    def get_connection(self, db_type: SqlStorageType, backend_name: str, is_async: bool) -> Optional[ConnectionData]:
        if ctx := self._ctx.get(None):
            return ctx.conns.get(EngineKey(db_type=db_type, backend=backend_name, is_async=is_async))
        return None

    def get_connections_for_type(self, db_type: SqlStorageType, is_async: bool) -> List[ConnectionData]:
        if ctx := self._ctx.get(None):
            backends = ctx.conns_by_type.get(db_type, set())
            conns = [self.get_connection(db_type, b, is_async) for b in backends]
            return [c for c in conns if c is not None]
        return []

    def get_connections_for_backend(self, backend_name: str, is_async: bool) -> List[ConnectionData]:
        if ctx := self._ctx.get(None):
            db_types = ctx.conns_by_backend.get(backend_name, set())
            conns = [self.get_connection(t, backend_name, is_async) for t in db_types]
            return [c for c in conns if c is not None]
        return []

    def get_or_create_connection(
        self, db_type: SqlStorageType, backend_name: str, is_async: bool,
            timeout_secs: Optional[int] = None) -> Tuple[bool, Optional[ConnectionData]]:
        if conn := self.get_connection(db_type, backend_name, is_async):
            return False, conn

        if engine := self._engine_factory.get_engine(db_type, backend_name, is_async):
            return True, ConnectionData(
                engine=engine, connection=engine.new_connection(),
                timeout_secs=timeout_secs or engine.timeout.total_seconds())

        return False, None

    def _add_connection(
            self, db_type: SqlStorageType, backend_name: str, is_async: bool, data: ConnectionData) -> ConnectionData:
        if ctx := self._ctx.get(None):
            ctx.conns[EngineKey(db_type=db_type, backend=backend_name, is_async=is_async)] = data
            ctx.conns_by_type[db_type].add(backend_name)
            ctx.conns_by_backend[backend_name].add(db_type)
        return data

    def _remove_connection(self, db_type: SqlStorageType, backend_name: str, is_async: bool):
        if ctx := self._ctx.get(None):
            ctx.conns.pop(EngineKey(db_type=db_type, backend=backend_name, is_async=is_async))
            ctx.conns_by_type[db_type].remove(backend_name)
            ctx.conns_by_backend[backend_name].remove(db_type)

    @contextlib.contextmanager
    def connection_context(
        self, db_type: SqlStorageType, backend_name: str,
            begin_tx: bool = True, autocommit: bool = True, timeout_secs: Optional[int] = None):
        created, data = self.get_or_create_connection(db_type, backend_name, False, timeout_secs)

        if not created:
            if begin_tx and not data.in_transaction:
                with data.engine.transaction_context(data.connection, autocommit):
                    yield self._add_connection(db_type, backend_name, False, replace(data, in_transaction=True))
                    self._add_connection(db_type, backend_name, False, replace(data, in_transaction=False))
            else:
                yield data
            return

        with data.connection as conn:
            assert isinstance(conn, data.engine.backend.connection_class), \
                f'Engine for DB "{db_type.value}" and backend "{backend_name}" has only async driver, ' \
                f'use async_connection_context() function instead'

            data = replace(data, connection=conn, in_transaction=begin_tx)
            if begin_tx:
                with data.engine.transaction_context(data.connection, autocommit):
                    yield self._add_connection(db_type, backend_name, False, data)
            else:
                yield self._add_connection(db_type, backend_name, False, data)
            self._remove_connection(db_type, backend_name, False)

    @contextlib.asynccontextmanager
    async def async_connection_context(
        self, db_type: SqlStorageType, backend_name: str,
            begin_tx: bool = True, autocommit: bool = True, timeout_secs: Optional[int] = None):
        created, data = self.get_or_create_connection(db_type, backend_name, True, timeout_secs)

        if not created:
            if begin_tx and not data.in_transaction:
                async with data.engine.async_transaction_context(data.connection, autocommit):
                    yield self._add_connection(db_type, backend_name, True, replace(data, in_transaction=True))
                    self._add_connection(db_type, backend_name, True, replace(data, in_transaction=False))
            else:
                yield data
            return

        async with data.connection as conn:
            assert isinstance(conn, data.engine.backend.async_connection_class), \
                f'Engine for DB "{db_type.value}" and backend "{backend_name}" has only sync driver, ' \
                f'use connection_context() function instead'

            data = replace(data, connection=conn, in_transaction=begin_tx)
            if begin_tx:
                async with data.engine.async_transaction_context(data.connection, autocommit):
                    yield self._add_connection(db_type, backend_name, True, data)
            else:
                yield self._add_connection(db_type, backend_name, True, data)
            self._remove_connection(db_type, backend_name, True)

    @contextlib.contextmanager
    def _with_engines_context(
            self, engines: Mapping[str, Sequence[SqlStorageEngine]], begin_tx: bool = True, autocommit: bool = True):
        ctx = ConnectionContext()
        token = self._ctx.set(ctx)
        try:
            with contextlib.ExitStack() as stack:
                for backend_name, engine_list in engines.items():
                    for engine in engine_list:
                        engine.start()
                        stack.callback(engine.stop)
                        stack.enter_context(self.connection_context(engine.db_type, backend_name, begin_tx, autocommit))
                yield ctx
        finally:
            if self._ctx.get(None) is ctx:
                self._ctx.reset(token)

    @contextlib.asynccontextmanager
    async def _with_async_engines_context(
            self, engines: Mapping[str, Sequence[SqlStorageEngine]], begin_tx: bool = True, autocommit: bool = True):
        ctx = ConnectionContext()
        token = self._ctx.set(ctx)
        try:
            async with contextlib.AsyncExitStack() as stack:
                for backend_name, engine_list in engines.items():
                    for engine in engine_list:
                        if engine.is_async:
                            await engine.async_start()
                            stack.push_async_callback(engine.async_stop)
                        else:
                            engine.start()
                            stack.callback(engine.stop)

                        await stack.enter_async_context(
                            self.async_connection_context(engine.db_type, backend_name, begin_tx, autocommit))
                yield ctx
        finally:
            if self._ctx.get(None) is ctx:
                self._ctx.reset(token)

    @contextlib.contextmanager
    def set_context(self, ctx: ConnectionContext):
        token = self._ctx.set(ctx)
        yield
        self._ctx.reset(token)

    def get_engines(self, backends: List[str], is_async: bool) -> Mapping[str, List[SqlStorageEngine]]:
        engines = defaultdict(list)

        if not backends:
            backends = self._engine_factory.get_backend_names()

        for backend_name in backends:
            for engine in self._engine_factory.get_engines_for_backend(backend_name, is_async):
                engines[backend_name].append(engine.engine)

        return engines

    @contextlib.contextmanager
    def with_engines(self, backends: List[str] = None, begin_tx: bool = True, autocommit: bool = True):
        engines = self.get_engines(backends or [], False)
        with self._with_engines_context(engines, begin_tx, autocommit) as ctx:
            yield ctx

    @contextlib.asynccontextmanager
    async def with_async_engines(self, backends: List[str] = None, begin_tx: bool = True, autocommit: bool = True):
        engines = self.get_engines(backends or [], True)
        async with self._with_async_engines_context(engines, begin_tx, autocommit) as ctx:
            yield ctx

    def use_engines(self, backends: List[str] = None, begin_tx: bool = True, autocommit: bool = True):
        def inner(function):
            if asyncio.iscoroutinefunction(function):
                engines = self.get_engines(backends or [], True)

                async def decorator(*args, **kwargs):
                    async with self._with_async_engines_context(engines, begin_tx, autocommit):
                        return await function(*args, **kwargs)

                return decorator
            else:
                engines = self.get_engines(backends or [], False)

                def decorator(*args, **kwargs):
                    with self._with_engines_context(engines, begin_tx, autocommit):
                        return function(*args, **kwargs)

                return decorator

        return inner



from fastapi import FastAPI

from dropland.log import logger, tr
from dropland.tasks import USE_SCHEDULER, Scheduler


def add_scheduler(app: FastAPI, scheduler: Scheduler):
    if not USE_SCHEDULER:
        return

    @app.on_event('startup')
    def init_task():
        app.state.scheduler = scheduler
        if not scheduler.running:
            scheduler.start()

        logger.info(tr('dropland.scheduler.started'))

    @app.on_event('shutdown')
    def fini_task():
        scheduler.shutdown(wait=True)
        app.state.scheduler = None

        logger.info(tr('dropland.scheduler.stopped'))



from datetime import datetime
from typing import Callable, Optional

import rpyc
from apscheduler.schedulers.base import STATE_STOPPED
from apscheduler.util import undefined
from dependency_injector.wiring import Provide, inject
from rpyc.utils.server import Server

from dropland.log import tr, worker_logger
from dropland.tasks.engine import EngineConfig
from dropland.tasks.local import Scheduler


class SchedulerService(rpyc.Service):
    def __init__(self, scheduler: Scheduler):
        self._scheduler = scheduler

    def _prepare_args(self, args: tuple) -> tuple:
        result = []
        for value in args:
            if isinstance(value, datetime):
                value = value.strftime('%Y-%m-%dT%H:%M:%S.%f')
            elif isinstance(value, type(undefined)):
                value = undefined
            result.append(value)
        return tuple(result)

    def _prepare_kwargs(self, args: dict) -> dict:
        for k, v in args.items():
            if isinstance(v, datetime):
                args[k] = v.strftime('%Y-%m-%dT%H:%M:%S.%f')
            elif isinstance(v, type(undefined)):
                args[k] = undefined
        return args

    def exposed_start(self, paused=False):
        return self._scheduler.start(paused)

    def exposed_shutdown(self, wait=True):
        return self._scheduler.shutdown(wait)

    def exposed_pause(self):
        return self._scheduler.pause()

    def exposed_resume(self):
        return self._scheduler.resume()

    def exposed_add_job(self, func, *args, **kwargs):
        return self._scheduler.add_job(func, *self._prepare_args(args), **self._prepare_kwargs(kwargs))

    def exposed_scheduled_job(self, trigger, *args, **kwargs):
        return self._scheduler.scheduled_job(trigger, *args, **kwargs)

    def exposed_modify_job(self, job_id, jobstore=None, **changes):
        if _ := self._scheduler.get_job(job_id, jobstore):
            self._scheduler.modify_job(job_id, jobstore, **changes)
            return True
        return False

    def exposed_reschedule_job(self, job_id, jobstore=None, trigger=None, **trigger_args):
        return self._scheduler.reschedule_job(job_id, jobstore, trigger, **trigger_args)

    def exposed_pause_job(self, job_id, jobstore=None):
        return self._scheduler.pause_job(job_id, jobstore)

    def exposed_resume_job(self, job_id, jobstore=None):
        return self._scheduler.resume_job(job_id, jobstore)

    def exposed_remove_job(self, job_id, jobstore=None):
        return self._scheduler.remove_job(job_id, jobstore)

    def exposed_remove_all_jobs(self, jobstore=None):
        return self._scheduler.remove_all_jobs(jobstore)

    def exposed_get_job(self, job_id, jobstore=None):
        return self._scheduler.get_job(job_id, jobstore)

    def exposed_get_jobs(self, jobstore=None, pending=None):
        return self._scheduler.get_jobs(jobstore, pending)

    @property
    def exposed_running(self):
        return self._scheduler.running


def service_main(scheduler: Scheduler, server: Server):
    try:
        server.start()

    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception:
        worker_logger.exception(tr('dropland.scheduler.service.exception'))
    finally:
        worker_logger.info(tr('dropland.scheduler.service.stop'))
        if STATE_STOPPED != scheduler.state:
            scheduler.shutdown()


@inject
def run_service(
        on_start: Optional[Callable[[Scheduler], None]] = None,
        on_shutdown: Optional[Callable[[Scheduler], None]] = None,
        engine_config=Provide['config.engine_config'],
        container=Provide['<container>']):
    """ Worker service entrypoint """
    from rpyc.utils.server import ThreadPoolServer

    worker_logger.info(tr('dropland.scheduler.service.start'))

    config = rpyc.core.protocol.DEFAULT_CONFIG.copy()
    config.update({'allow_public_attrs': True, 'allow_pickle': True, 'import_custom_exceptions': True})

    container.is_scheduler.override(True)
    container.reset_singletons()
    scheduler = container.instance()

    engine_config = engine_config if isinstance(engine_config, EngineConfig) else EngineConfig(**engine_config)

    server = ThreadPoolServer(
        SchedulerService(scheduler), nbThreads=engine_config.task_workers,
        hostname=engine_config.task_host, port=engine_config.task_port, protocol_config=config)

    if on_start:
        on_start(scheduler)

    service_main(scheduler, server)

    if on_shutdown:
        on_shutdown(scheduler)


if __name__ == '__main__':
    run_service()



import pytest

from tests.sql_models_data import USE_GINO

pytestmark = pytest.mark.skipif(not USE_GINO, reason='For Gino only')
gino = pytest.importorskip('gino')

if USE_GINO:
    from dropland.storages.containers import default_sql_storage
    from dropland.storages.sql import EngineConfig, SqlStorageType
    from tests import MYSQL_URI, POSTGRES_URI


@pytest.mark.asyncio
async def test_create_engine():
    pg_config = EngineConfig(
        url=POSTGRES_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )
    mysql_config = EngineConfig(
        url=MYSQL_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine_factory = default_sql_storage.engine_factory()

    assert not engine_factory.create_engine(SqlStorageType.SQLITE, pg_config, use_async=False)
    assert not engine_factory.create_engine(SqlStorageType.SQLITE, pg_config, use_async=True)
    assert not engine_factory.create_engine(SqlStorageType.POSTGRES, pg_config, use_async=False)
    assert engine_factory.create_engine(SqlStorageType.POSTGRES, pg_config, use_async=True)
    assert not engine_factory.create_engine(SqlStorageType.MYSQL, mysql_config, use_async=False)
    assert engine_factory.create_engine(SqlStorageType.MYSQL, mysql_config, use_async=True)

    for db_type in (SqlStorageType.SQLITE, SqlStorageType.POSTGRES, SqlStorageType.MYSQL):
        for e in engine_factory.get_engines_for_type(db_type, is_async=True):
            assert e.db_type == db_type
            assert e.engine
            assert e.engine.is_async

    for e in engine_factory.get_engines_for_backend('gino', is_async=True):
        assert e.backend == 'gino'
        assert e.engine

    assert engine_factory.get_backend_names() == ['gino']


@pytest.mark.asyncio
async def test_create_pg_engine():
    config = EngineConfig(
        url=POSTGRES_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine = default_sql_storage.create_engine(SqlStorageType.POSTGRES, config, use_async=True)
    assert engine

    with pytest.raises(RuntimeError):
        engine.start()

    await engine.async_start()

    async with engine.new_connection() as conn:
        res = await conn.scalar('select version();')
        print(res)
        res = await conn.scalar('select 1 + 2;')
        assert res == 3

    with pytest.raises(RuntimeError):
        engine.stop()

    await engine.async_stop()

    with pytest.raises(gino.exceptions.UninitializedError):
        assert engine.new_connection()


@pytest.mark.asyncio
async def test_create_mysql_engine():
    config = EngineConfig(
        url=MYSQL_URI,
        pool_min_size=1, pool_max_size=2,
        pool_expire_seconds=15, pool_timeout_seconds=15,
    )

    engine = default_sql_storage.create_engine(SqlStorageType.MYSQL, config, use_async=True)
    assert engine

    with pytest.raises(RuntimeError):
        engine.start()

    await engine.async_start()

    async with engine.new_connection() as conn:
        res = await conn.scalar('select version();')
        print(res)
        res = await conn.scalar('select 1 + 2;')
        assert res == 3

    await engine.async_stop()

    with pytest.raises(gino.exceptions.UninitializedError):
        assert engine.new_connection()



import asyncio
import contextlib
from dataclasses import dataclass, replace
from typing import Dict, List, Mapping, Optional, Tuple

from aioredis import Redis as RedisConnection
from contextvars import ContextVar

from dropland.storages.redis.engine import RedisStorageEngine, RedisStorageBackend


@dataclass
class ConnectionData:
    engine: RedisStorageEngine
    connection: RedisConnection


class ConnectionContext:
    def __init__(self):
        self.conns: Dict[str, ConnectionData] = dict()


class ConnectionManager:
    def __init__(self, engine_factory: RedisStorageBackend):
        self._ctx: ContextVar[ConnectionContext] = ContextVar('_ctx')
        self._engine_factory = engine_factory

    def get_connection(self, name: str) -> Optional[ConnectionData]:
        if ctx := self._ctx.get(None):
            return ctx.conns.get(name)
        return None

    def get_or_create_connection(self, name: str) -> Tuple[bool, Optional[ConnectionData]]:
        if conn := self.get_connection(name):
            return False, conn

        if engine := self._engine_factory.get_engine(name):
            return True, ConnectionData(
                engine=engine,
                connection=engine.new_connection())

        return False, None

    def _add_connection(self, name: str, data: ConnectionData) -> ConnectionData:
        if ctx := self._ctx.get(None):
            ctx.conns[name] = data
        return data

    def _remove_connection(self, name: str):
        if ctx := self._ctx.get(None):
            ctx.conns.pop(name)

    @contextlib.asynccontextmanager
    async def async_connection_context(self, name: str):
        created, data = self.get_or_create_connection(name)

        if not created:
            yield data
            return

        async with data.connection as conn:
            data = replace(data, connection=conn)
            yield self._add_connection(name, data)
            self._remove_connection(name)

    @contextlib.asynccontextmanager
    async def _with_async_engines_context(self, engines: Mapping[str, RedisStorageEngine]):
        ctx = ConnectionContext()
        token = self._ctx.set(ctx)
        try:
            async with contextlib.AsyncExitStack() as stack:
                for name, engine in engines.items():
                    assert engine is self._engine_factory.get_engine(name)
                    if engine.is_async:
                        await engine.async_start()
                        stack.push_async_callback(engine.async_stop)
                    else:
                        engine.start()
                        stack.callback(engine.stop)

                    await stack.enter_async_context(self.async_connection_context(name))
                yield ctx
        finally:
            if self._ctx.get(None) is ctx:
                self._ctx.reset(token)

    @contextlib.contextmanager
    def set_context(self, ctx: ConnectionContext):
        token = self._ctx.set(ctx)
        yield
        self._ctx.reset(token)

    def get_engines(self, names: List[str]) -> Mapping[str, RedisStorageEngine]:
        engines = dict()

        if not names:
            names = self._engine_factory.get_engine_names()

        for name in names:
            if engine := self._engine_factory.get_engine(name):
                engines[name] = engine

        return engines

    @contextlib.asynccontextmanager
    async def with_async_engines(self, names: List[str] = None):
        engines = self.get_engines(names or [])
        async with self._with_async_engines_context(engines) as ctx:
            yield ctx

    def use_engines(self, names: List[str] = None):
        def inner(function):
            assert asyncio.iscoroutinefunction(function)
            engines = self.get_engines(names or [])

            async def decorator(*args, **kwargs):
                async with self._with_async_engines_context(engines):
                    return await function(*args, **kwargs)

            return decorator
        return inner



import pytest

from tests.sql_models_data import USE_SQLA

pytestmark = pytest.mark.skipif(not USE_SQLA, reason='For SqlAlchemy only')

if USE_SQLA:
    from dropland.storages.sql import SqlStorageType


@pytest.mark.asyncio
async def test_init_context(test_sql_storage, sqlite_engine, sqlite_async_engine):
    mng = test_sql_storage.manager()

    @mng.use_engines(['sqla'])
    def sync_context_func():
        assert mng.get_connection(SqlStorageType.SQLITE, 'sqla', False)
        assert not mng.get_connection(SqlStorageType.SQLITE, 'sqla', True)

        for c in mng.get_connections_for_type(SqlStorageType.SQLITE, False):
            assert c.engine
            assert c.connection
            assert c.in_transaction

        assert 0 == len(mng.get_connections_for_type(SqlStorageType.SQLITE, True))

        for c in mng.get_connections_for_backend('sqla', False):
            assert c.engine
            assert c.connection
            assert c.in_transaction

        assert 0 == len(mng.get_connections_for_backend('sqla', True))

    @mng.use_engines(['sqla'])
    async def async_context_func():
        assert not mng.get_connection(SqlStorageType.SQLITE, 'sqla', False)
        assert mng.get_connection(SqlStorageType.SQLITE, 'sqla', True)

        assert 0 == len(mng.get_connections_for_type(SqlStorageType.SQLITE, False))

        for c in mng.get_connections_for_type(SqlStorageType.SQLITE, True):
            assert c.engine
            assert c.connection
            assert c.in_transaction

        assert 0 == len(mng.get_connections_for_backend('sqla', False))

        for c in mng.get_connections_for_backend('sqla', True):
            assert c.engine
            assert c.connection
            assert c.in_transaction

    sync_context_func()
    await async_context_func()


@pytest.mark.asyncio
async def test_connections(test_sql_storage, sqlite_engine, sqlite_async_engine):
    mng = test_sql_storage.manager()

    @mng.use_engines(['sqla'])
    def sync_context_func():
        created, sql_data = mng.get_or_create_connection(SqlStorageType.SQLITE, 'sqla', False)
        assert not created
        assert sql_data

        assert sql_data.engine == sqlite_engine

        created, sql_data = mng.get_or_create_connection(SqlStorageType.SQLITE, 'sqla', True)
        assert created
        assert sql_data

        assert sql_data.engine == sqlite_async_engine

    @mng.use_engines(['sqla'])
    async def async_context_func():
        created, sql_data = mng.get_or_create_connection(SqlStorageType.SQLITE, 'sqla', False)
        assert created
        assert sql_data

        assert sql_data.engine == sqlite_engine

        created, sql_data = mng.get_or_create_connection(SqlStorageType.SQLITE, 'sqla', True)
        assert not created
        assert sql_data

        assert sql_data.engine == sqlite_async_engine

    sync_context_func()
    await async_context_func()



from datetime import datetime
from random import randint

import pytest
from sqlalchemy import Column, DateTime, Integer, String, func, text

from dropland.storages.redis import USE_REDIS
from dropland.storages.sql import USE_SQL

pytestmark = pytest.mark.skipif(not USE_SQL, reason='For Sql only')

if USE_REDIS:
    from dropland.data.model import CachedModel
    from .redis_models_data import redis_engine

if USE_SQL:
    from .sql_models_data import USE_GINO, USE_SQLA, \
        gino_mysql, gino_pg, sqla_mysql, sqla_pg, sqla_sqlite

    sqla_engines = (sqla_sqlite, sqla_pg, sqla_mysql)
    gino_engines = (gino_pg, gino_mysql)

    engines = sqla_engines if USE_SQLA else gino_engines
    models = [e.Model for e in engines]
    ids = ['SQLite', 'PostgreSQL', 'MySQL'] if USE_SQLA else ['PostgreSQL', 'MySQL']

    if USE_REDIS:
        models += [CachedModel('dropland', e.Model, redis_engine.Model) for e in engines]
        ids += [f'{i}+Redis' for i in ids]
else:
    models, ids = [], []


# noinspection PyPep8Naming
@pytest.mark.asyncio
@pytest.fixture(scope='module', params=models, ids=ids)
async def TestModel(request):
    Model = request.param
    engine = request.param.get_engine()
    table_name = f'testmodel_{ids[request.param_index].lower()}'

    class SomeModel(Model):
        __tablename__ = table_name

        id = Column(Integer, primary_key=True)
        data = Column(String(128))
        create_date = Column(DateTime, server_default=func.now())

        # required in order to access columns with server defaults
        # or SQL expression defaults, subsequent to a flush, without
        # triggering an expired load
        __mapper_args__ = {'eager_defaults': True}

    assert SomeModel.get_engine() is engine
    ti = SomeModel.get_table_info()
    assert ti.backend == engine.backend.name
    assert ti.db_type == engine.db_type
    assert ti.is_async == engine.is_async

    if USE_SQLA:
        async with engine.raw_engine.begin() as conn:
            await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
            await conn.run_sync(engine.metadata.create_all, tables=[SomeModel.__table__])
            await conn.commit()

        yield SomeModel

        async with engine.raw_engine.begin() as conn:
            await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
            await conn.commit()

    elif USE_GINO:
        from dropland.storages.sql import default_storage
        from gino.schema import GinoSchemaVisitor

        async with default_storage.manager().with_async_engines(['gino']):
            bind = engine.raw_engine.bind
            await GinoSchemaVisitor(engine.metadata).drop_all(bind, tables=[SomeModel.__table__])
            await GinoSchemaVisitor(engine.metadata).create_all(bind, tables=[SomeModel.__table__])

        yield SomeModel

        async with default_storage.manager().with_async_engines(['gino']):
            bind = engine.raw_engine.bind
            await GinoSchemaVisitor(engine.metadata).drop_all(bind, tables=[SomeModel.__table__])


_construct_counter = 0
_construct_rela_counter = 0
_construct_list_counter = 0
_construct_rela_list_counter = 0
_reload_rela_counter = 0


@pytest.fixture(scope='function')
def build_calls_counter(TestModel, monkeypatch):
    global _construct_counter, _construct_list_counter, \
        _construct_rela_counter, _construct_rela_list_counter, _reload_rela_counter
    _construct_counter = 0
    _construct_list_counter = 0
    _construct_rela_counter = 0
    _construct_rela_list_counter = 0
    _reload_rela_counter = 0

    old_construct = TestModel.construct
    old_construct_rela = TestModel.construct_rela
    old_construct_list = TestModel.construct_list
    old_construct_rela_list = TestModel.construct_rela_list
    old_reload_rela = TestModel.reload_rela

    async def construct(context, data, **kwargs):
        global _construct_counter
        _construct_counter += 1
        return await old_construct(context, data, **kwargs)

    async def construct_rela(context, instance, **kwargs):
        global _construct_rela_counter
        _construct_rela_counter += 1
        return await old_construct_rela(context, instance, **kwargs)

    async def construct_list(context, objects, **kwargs):
        global _construct_list_counter
        _construct_list_counter += 1
        return await old_construct_list(context, objects, **kwargs)

    async def construct_rela_list(context, objects, **kwargs):
        global _construct_rela_list_counter
        _construct_rela_list_counter += 1
        return await old_construct_rela_list(context, objects, **kwargs)

    async def reload_rela(self, context, load_fields, **kwargs):
        global _reload_rela_counter
        _reload_rela_counter += 1
        return await old_reload_rela(self, context, load_fields, **kwargs)

    monkeypatch.setattr(TestModel, 'construct', construct)
    monkeypatch.setattr(TestModel, 'construct_rela', construct_rela)
    monkeypatch.setattr(TestModel, 'construct_list', construct_list)
    monkeypatch.setattr(TestModel, 'construct_rela_list', construct_rela_list)
    monkeypatch.setattr(TestModel, 'reload_rela', reload_rela)


@pytest.mark.asyncio
async def test_model_fields(data_ctx, TestModel):
    m = TestModel()
    assert TestModel.get_fields() == {'id', 'create_date', 'data'}
    assert m.get_values() == {'id': None, 'create_date': None, 'data': None}

    m.id = 123
    m.data = 'qwer'
    assert m.get_values() == {'id': 123, 'create_date': None, 'data': 'qwer'}


@pytest.mark.asyncio
async def test_create(data_ctx, TestModel, build_calls_counter):
    obj = await TestModel.create(dict(data='123'))
    assert obj.id > 0
    assert obj.data == '123'

    assert _construct_counter == 1
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_get_or_create(data_ctx, TestModel, build_calls_counter):
    obj1, created = await TestModel.get_or_create(999, dict(id=999, data='123'))
    assert created is True
    assert obj1.id == 999
    assert obj1.data == '123'

    obj2, created = await TestModel.get_or_create(999, dict(id=999, data='123'))
    assert created is False
    assert obj2.id == 999
    assert obj2.data == '123'

    assert _construct_counter == (1 if TestModel.has_cache() else 2)
    assert _construct_rela_counter == 2
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_rela_counter


@pytest.mark.asyncio
async def test_update(data_ctx, TestModel, build_calls_counter):
    obj = await TestModel.create(dict(data='123'))
    assert await obj.update(dict(data='456'))
    assert obj.id > 0
    assert obj.data == '456'

    assert _construct_counter == 1
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter + 1


@pytest.mark.asyncio
async def test_update_by_id(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='123'))

    obj2 = await TestModel.update_by_id(obj1.id, dict(data='234'))
    assert obj2.id > 0
    assert obj2.data == '234'

    obj3 = await TestModel.update_by_id(-1, dict(data='234'))
    assert obj3 is None

    assert _construct_counter == 2
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_update_by(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.update_by([TestModel.data != '1'], dict(data='0'))
    assert res == 2

    assert obj1.data == '1'
    assert obj2.data == '0'
    assert obj3.data == '0'

    assert _construct_counter == 3
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_get(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='123'))

    obj2 = await TestModel.get(obj1.id)
    assert obj1.id == obj2.id
    assert obj1.data == obj2.data
    assert obj1 == obj2

    obj3 = await TestModel.get(-1)
    assert obj3 is None

    assert _construct_counter == (1 if TestModel.has_cache() else 2)
    assert _construct_rela_counter == 2
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_rela_counter


@pytest.mark.asyncio
async def test_list(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    objects = await TestModel.list()
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list(skip=1)
    assert len(objects) == 2
    assert objects[0] == obj2
    assert objects[1] == obj3

    objects = await TestModel.list(skip=2)
    assert len(objects) == 1
    assert objects[0] == obj3

    objects = await TestModel.list(skip=3)
    assert len(objects) == 0

    objects = await TestModel.list(limit=1)
    assert len(objects) == 1
    assert objects[0] == obj1

    objects = await TestModel.list(limit=2)
    assert len(objects) == 2
    assert objects[0] == obj1
    assert objects[1] == obj2

    objects = await TestModel.list(limit=3)
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list(skip=1, limit=3)
    assert len(objects) == 2
    assert objects[0] == obj2
    assert objects[1] == obj3

    objects = await TestModel.list(skip=2, limit=1)
    assert len(objects) == 1
    assert objects[0] == obj3

    objects = await TestModel.list(skip=3, limit=3)
    assert len(objects) == 0

    objects = await TestModel.list(skip=2, limit=0)
    assert len(objects) == 1
    assert objects[0] == obj3

    assert _construct_counter == 19
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 11
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_filter(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    objects = await TestModel.list()
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list([])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.list([TestModel.data == '-1'])
    assert len(objects) == 0

    objects = await TestModel.list([TestModel.data == '1'])
    assert len(objects) == 1
    assert objects[0] == obj1

    objects = await TestModel.list([TestModel.data == '1', TestModel.data == '2'])
    assert len(objects) == 0

    objects = await TestModel.list([(TestModel.data == '1') | (TestModel.data == '2')])
    assert len(objects) == 2
    assert objects[0] == obj1
    assert objects[1] == obj2

    objects = await TestModel.list([(TestModel.data == '1') & (TestModel.id == obj1.id)])
    assert len(objects) == 1
    assert objects[0] == obj1

    objects = await TestModel.list([text('data > :data')], params=dict(data='1'))
    assert len(objects) == 2
    assert objects[0] == obj2
    assert objects[1] == obj3

    assert _construct_counter == 15
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 8
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_any(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    objects = await TestModel.get_any([])
    assert len(objects) == 0

    objects = await TestModel.get_any([obj1.id, obj2.id, obj3.id])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3

    objects = await TestModel.get_any([obj1.id, obj3.id])
    assert len(objects) == 2
    assert objects[0] == obj1
    assert objects[1] == obj3

    objects = await TestModel.get_any([-1, obj1.id, 999999, obj3.id, 0])
    assert len(objects) == 5
    assert objects[0] is None
    assert objects[1] == obj1
    assert objects[2] is None
    assert objects[3] == obj3
    assert objects[4] is None

    assert _construct_counter == 10
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == (5 if TestModel.has_cache() else 4)
    assert _construct_rela_list_counter == 4
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_count(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.count()
    assert res == 3

    res = await TestModel.count([])
    assert res == 3

    res = await TestModel.count([TestModel.data == '-1'])
    assert res == 0

    res = await TestModel.count([TestModel.data == '1'])
    assert res == 1

    res = await TestModel.count([TestModel.data == '1', TestModel.data == '2'])
    assert res == 0

    res = await TestModel.count([(TestModel.data == '1') | (TestModel.data == '2')])
    assert res == 2

    res = await TestModel.count([(TestModel.data == '1') & (TestModel.id == obj1.id)])
    assert res == 1

    res = await TestModel.count([text('data > :data')], params=dict(data='1'))
    assert res == 2

    assert _construct_counter == 3
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_exists(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.exists(obj1.id)
    assert res is True

    res = await TestModel.exists(-1)
    assert res is False

    res = await TestModel.exists_by([])
    assert res is False

    res = await TestModel.exists_by([TestModel.data == '-1'])
    assert res is False

    res = await TestModel.exists_by([TestModel.data == '1'])
    assert res is True

    res = await TestModel.exists_by([TestModel.data == '1', TestModel.data == '2'])
    assert res is False

    res = await TestModel.exists_by([(TestModel.data == '1') | (TestModel.data == '2')])
    assert res is True

    res = await TestModel.exists_by([(TestModel.data == '1') & (TestModel.id == obj1.id)])
    assert res is True

    res = await TestModel.exists_by([text('data > :data')], params=dict(data='1'))
    assert res is True

    res = await TestModel.exists_by([text('data > :data')], params=dict(data='3'))
    assert res is False

    assert _construct_counter == 3
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter


@pytest.mark.asyncio
async def test_delete(data_ctx, TestModel, build_calls_counter):
    obj = await TestModel.create(dict(data='123'))

    objects = await TestModel.list()
    assert len(objects) == 1
    assert objects[0] == obj
    obj2 = await TestModel.get(obj.id)
    assert obj == obj2

    assert (await obj.delete()) is True

    objects = await TestModel.list()
    assert len(objects) == 0
    obj2 = await TestModel.get(obj.id)
    assert obj2 is None

    assert (await obj.delete()) is False

    assert _construct_counter == (2 if TestModel.has_cache() else 3)
    assert _construct_rela_counter == 3
    assert _construct_list_counter == 2
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_rela_counter


@pytest.mark.asyncio
async def test_delete_by_id(data_ctx, TestModel, build_calls_counter):
    obj = await TestModel.create(dict(data='123'))

    objects = await TestModel.list()
    assert len(objects) == 1
    assert objects[0] == obj
    obj2 = await TestModel.get(obj.id)
    assert obj == obj2

    assert (await TestModel.delete_by_id(obj.id)) is True

    objects = await TestModel.list()
    assert len(objects) == 0
    obj2 = await TestModel.get(obj.id)
    assert obj2 is None

    assert (await TestModel.delete_by_id(obj.id)) is False
    assert (await TestModel.delete_by_id(-1)) is False

    assert _construct_counter == (2 if TestModel.has_cache() else 3)
    assert _construct_rela_counter == 3
    assert _construct_list_counter == 2
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == 3


@pytest.mark.asyncio
async def test_delete_by(data_ctx, TestModel, build_calls_counter):
    obj1 = await TestModel.create(dict(data='1'))
    obj2 = await TestModel.create(dict(data='2'))
    obj3 = await TestModel.create(dict(data='3'))

    res = await TestModel.delete_by([TestModel.data != '1'])
    assert res == 2

    res = await TestModel.delete_by([TestModel.data != '1'])
    assert res == 0

    assert await obj1.load()
    assert not await obj2.load()
    assert not await obj3.load()

    assert _construct_counter == 3
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter + 1


@pytest.mark.asyncio
async def test_save_and_load(data_ctx, TestModel, build_calls_counter):
    m = TestModel()
    m.id = randint(1, 1000000)
    m.data = 'qwer'
    assert m.create_date is None

    assert not await TestModel.get(m.id)

    assert await m.save()
    assert isinstance(m.create_date, datetime)

    m2 = await TestModel.get(m.id)
    assert m2.id == m.id
    assert m2.data == m.data
    assert isinstance(m2.create_date, datetime)
    assert m2.create_date == m.create_date

    m2.id = 1234
    m2.data = 'qwerty'
    m2.create_date = datetime.now()
    assert await m2.save(updated_fields=['id', 'create_date'])

    m3 = await TestModel.get(1234)
    assert m3.id == 1234
    assert m3.data == 'qwer'
    assert m3.create_date.replace(microsecond=0) == m2.create_date.replace(microsecond=0)

    assert await m2.load()
    assert m2.id == 1234
    assert m2.data == 'qwer'

    assert _construct_counter == (1 if TestModel.has_cache() else 2)
    assert _construct_rela_counter == 2
    assert _construct_list_counter == 0
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_rela_counter + 3


@pytest.mark.asyncio
async def test_save_all(data_ctx, TestModel, build_calls_counter):
    obj1 = TestModel(data='1')
    obj2 = TestModel(data='2')
    obj3 = TestModel(data='3')

    assert await TestModel.save_all([obj1, obj2, obj3])

    objects = await TestModel.get_any([obj1.id, obj2.id, obj3.id])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3
    assert objects[0].id == obj1.id
    assert objects[0].data == obj1.data
    assert objects[1].id == obj2.id
    assert objects[1].data == obj2.data
    assert objects[2].id == obj3.id
    assert objects[2].data == obj3.data

    obj1.data = '4'
    obj2.data = '5'
    obj3.data = '6'

    assert await TestModel.save_all([obj1, obj2, obj3])

    objects = await TestModel.get_any([obj1.id, obj2.id, obj3.id])
    assert len(objects) == 3
    assert objects[0] == obj1
    assert objects[1] == obj2
    assert objects[2] == obj3
    assert objects[0].id == obj1.id
    assert objects[0].data == obj1.data
    assert objects[1].id == obj2.id
    assert objects[1].data == obj2.data
    assert objects[2].id == obj3.id
    assert objects[2].data == obj3.data

    assert _construct_counter == 6
    assert _construct_rela_counter == _construct_counter
    assert _construct_list_counter == 2
    assert _construct_rela_list_counter == _construct_list_counter
    assert _reload_rela_counter == _construct_counter



from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Dict, List, Optional, Set

from dependency_injector.wiring import inject

from dropland.storages.sql.base import EngineConfig, SqlStorageEngine, SqlStorageType
from dropland.storages.sql.loader import SqlBackendLoader


@dataclass
class EngineKey:
    db_type: SqlStorageType
    backend: str
    is_async: bool

    def __eq__(self, other: 'EngineKey'):
        return (self.db_type, self.backend, self.is_async) == (other.db_type, other.backend, other.is_async)

    def __hash__(self):
        return hash((self.db_type, self.backend, self.is_async))


@dataclass
class EngineWithDbType:
    db_type: SqlStorageType
    engine: SqlStorageEngine


@dataclass
class EngineWithBackend:
    backend: str
    engine: SqlStorageEngine


class EngineFactory:
    def __init__(self, backend_loader: SqlBackendLoader):
        self._backend_loader = backend_loader
        self._engines: Dict[EngineKey, SqlStorageEngine] = dict()
        self._engines_by_type: Dict[SqlStorageType, Set[str]] = defaultdict(set)
        self._engines_by_backend: Dict[str, Set[SqlStorageType]] = defaultdict(set)

    @inject
    def create_engine(
        self, db_type: SqlStorageType, config: EngineConfig,
            backend_name: Optional[str] = None, use_async: bool = False) -> Optional[SqlStorageEngine]:
        uri = config.url

        if use_async:
            if db_type == SqlStorageType.SQLITE:
                uri = uri.replace('sqlite', 'sqlite+aiosqlite', 1)
            elif db_type == SqlStorageType.POSTGRES:
                uri = uri.replace('postgresql', 'postgresql+asyncpg', 1)
            elif db_type == SqlStorageType.MYSQL:
                uri = uri.replace('mysql', 'mysql+aiomysql', 1)
        else:
            if db_type == SqlStorageType.MYSQL:
                uri = uri.replace('mysql', 'mysql+pymysql', 1)

        config = replace(config, url=uri)
        self._backend_loader.load_backends()

        backends = self._backend_loader.get_backends_for_db(db_type)
        if len(backends) < 1:
            return None

        backend = None

        if backend_name:
            for b in backends:
                if b == backend_name:
                    backend = self._backend_loader.get_backend(b)
                    break
        else:
            backend_name = backends[0]
            backend = self._backend_loader.get_backend(backend_name)

        if not backend:
            return None

        if engine := self._engines.get(EngineKey(db_type=db_type, backend=backend_name, is_async=use_async)):
            return engine
        else:
            engine = backend.create_engine(db_type, config, use_async)

        if not engine:
            return None

        self._engines[EngineKey(db_type=db_type, backend=backend_name, is_async=use_async)] = engine
        self._engines_by_type[db_type].add(backend_name)
        self._engines_by_backend[backend_name].add(db_type)
        return engine

    def get_engine(self, db_type: SqlStorageType, backend_name: str, is_async: bool = False) \
            -> Optional[SqlStorageEngine]:
        return self._engines.get(EngineKey(db_type=db_type, backend=backend_name, is_async=is_async))

    def get_engines_for_type(self, db_type: SqlStorageType, is_async: bool) -> List[EngineWithDbType]:
        backends = self._engines_by_type.get(db_type, set())
        engines = [self.get_engine(db_type, b, is_async) for b in backends]
        return [EngineWithDbType(db_type=db_type, engine=e) for e in engines if e is not None]

    def get_engines_for_backend(self, backend_name: str, is_async: bool) -> List[EngineWithBackend]:
        db_types = self._engines_by_backend.get(backend_name, set())
        engines = [self.get_engine(t, backend_name, is_async) for t in db_types]
        return [EngineWithBackend(backend=backend_name, engine=e) for e in engines if e is not None]

    def get_backend_names(self) -> List[str]:
        return list(self._engines_by_backend.keys())



try:
    import starlette

    from .model import ApiModelMixin
    from .middleware import add_middleware
    from .permissions import HttpPermission, PermDepends, Authenticated, Admin
    from .scheduler import add_scheduler

except ImportError:
    pass



from datetime import timedelta, tzinfo
from typing import Any, Dict, Union

import pytz
from pydantic import BaseSettings, validator
from pytimeparse.timeparse import timeparse


class Settings(BaseSettings):
    TIMEZONE: tzinfo = pytz.UTC
    DEBUG: bool = False

    DATABASE_ECHO: bool = True
    DATABASE_POOL_MIN_SIZE: int = 1
    DATABASE_POOL_MAX_SIZE: int = 8
    DATABASE_POOL_EXPIRE_SECONDS: int = 60
    DATABASE_POOL_TIMEOUT_SECONDS: int = 15

    DEFAULT_CACHE_TTL: timedelta = timedelta(seconds=timeparse('1 min'))

    @validator('TIMEZONE', pre=True)
    def set_timezone(cls, v: Union[str, tzinfo], values: Dict[str, Any]) -> tzinfo:
        if isinstance(v, tzinfo):
            return v
        return pytz.timezone(v)

    @validator('DATABASE_ECHO', pre=True)
    def set_db_echo(cls, v: bool, values: Dict[str, Any]) -> bool:
        return values['DEBUG']

    @validator('DEFAULT_CACHE_TTL', pre=True)
    def set_default_ttl(cls, v: Union[str, timedelta], values: Dict[str, Any]) -> timedelta:
        if isinstance(v, timedelta):
            return v
        return timedelta(seconds=timeparse(v))

    class Config:
        case_sensitive = False
        env_file = '.env'



from typing import Any, Dict, Union

from starlette.exceptions import HTTPException
from starlette.status import HTTP_404_NOT_FOUND

from dropland.storages.sql.model import UpdateSchemaType


# noinspection PyUnresolvedReferences
class ApiModelMixin:
    @classmethod
    async def get_or_404(cls, id_value: Any, **kwargs) -> 'Model':
        res = await cls.get(id_value, **kwargs)
        if not res:
            raise HTTPException(HTTP_404_NOT_FOUND, 'Not found')
        return res

    @classmethod
    async def update_or_404(cls, id_value: Any, data: Union[UpdateSchemaType, Dict[str, Any]]) -> 'Model':
        res = await cls.update_by_id(id_value, data)
        if not res:
            raise HTTPException(HTTP_404_NOT_FOUND, 'Not found')
        return res

    @classmethod
    async def delete_or_404(cls, id_value: Any) -> bool:
        res = await cls.delete_by_id(id_value)
        if not res:
            raise HTTPException(HTTP_404_NOT_FOUND, 'Not found')
        return res



import contextlib
import pickle
from collections import OrderedDict
from datetime import timedelta
from typing import Any, Dict, List, Optional, Tuple

from dependency_injector.wiring import Provide, inject

from dropland.data.context import ContextData
from dropland.storages.base import CacheModel
from dropland.storages.redis import RedisStorageEngine
from dropland.storages.redis.connections import ConnectionManager


class RedisModel(type):
    def __init__(cls, class_name, parents, attributes):
        super().__init__(class_name, parents, attributes)

    @inject
    def __new__(mcs, class_name, parents, attributes, manager: ConnectionManager = Provide['manager']):
        rv = super().__new__(mcs, class_name, parents, attributes)
        if isinstance(manager, ConnectionManager):
            rv._redis_manager = manager
        return rv

    # noinspection PyUnresolvedReferences
    @property
    def _model_cache_key(cls):
        return f'{cls._redis_engine.name}.models.{cls.__name__}'


class RedisModelBase(CacheModel, metaclass=RedisModel):
    __metaclass__ = RedisModel
    _redis_engine: RedisStorageEngine = None
    _redis_manager: ConnectionManager = None

    @classmethod
    def get_model_cache_key(cls) -> str:
        return cls._model_cache_key

    @classmethod
    def get_engine(cls) -> RedisStorageEngine:
        return cls._redis_engine

    @classmethod
    @contextlib.asynccontextmanager
    async def _async_connection_context(cls, ctx: ContextData = None):
        ctx = ctx if ctx else ContextData()
        if hasattr(ctx, 'redis'):
            yield ctx
        else:
            async with cls._redis_manager.async_connection_context(cls._redis_engine.name) as c:
                ctx.redis = c
                yield ctx
                del ctx.redis


class SimpleRedisModelBase(RedisModelBase):
    @classmethod
    async def _cache_one(
        cls, ctx: ContextData, instance: Optional['RedisModelBase'] = None,
            id_value: Optional[Any] = None, data: Optional[Dict[str, Any]] = None,
            exp: Optional[timedelta] = None, **kwargs) -> bool:
        cache_kwargs = dict()
        if instance and id_value is None:
            id_value = instance.get_id_value()

        if cls.Meta.cache_ttl_enable:
            expiration = exp or ctx.data.get(f'exp/{cls.get_model_cache_key()}', ctx.redis.engine.default_ttl)
            total_seconds = expiration.total_seconds()

            if total_seconds > 1:
                cache_kwargs['expire'] = int(total_seconds)
            elif 0 < total_seconds < 1:
                cache_kwargs['pexpire'] = int(total_seconds * 1000)
            else:
                total_seconds = ctx.redis.engine.default_ttl.total_seconds()
                cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

        if data:
            data = pickle.dumps(data)
        else:
            data = instance.serialize() if instance else pickle.dumps(None)

        return bool(await ctx.redis.connection.set(cls.get_cache_key(id_value), data, **cache_kwargs))

    @classmethod
    async def _cache_many(
            cls, ctx: ContextData, objects: List['RedisModelBase'], exp: timedelta = None, **kwargs) -> bool:
        if not objects:
            return False
        model_cache_key = cls.get_model_cache_key()
        res = False

        tx = ctx.redis.connection.multi_exec()

        for instance in objects:
            cache_kwargs = dict()

            if cls.Meta.cache_ttl_enable:
                expiration = exp or ctx.data.get(f'exp/{model_cache_key}', ctx.redis.engine.default_ttl)
                total_seconds = expiration.total_seconds()

                if total_seconds > 1:
                    cache_kwargs['expire'] = int(total_seconds)
                elif 0 < total_seconds < 1:
                    cache_kwargs['pexpire'] = int(total_seconds * 1000)
                else:
                    total_seconds = ctx.redis.engine.default_ttl.total_seconds()
                    cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

            tx.set(cls.get_cache_key(instance.get_id_value()), instance.serialize(), **cache_kwargs)

        for r in await tx.execute():
            res |= r

        return bool(res)

    @classmethod
    async def _load_one(
            cls, ctx: ContextData, cache_key: str, **kwargs) -> Tuple[bool, Optional[Dict[str, Any]]]:
        if res := await ctx.redis.connection.get(cache_key):
            return True, cls.deserialize(res)
        return False, None

    @classmethod
    async def _load_many(
            cls, ctx: ContextData, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        if not indices:
            return []
        cache_keys = [cls.get_cache_key(id_value) for id_value in indices]
        res = await ctx.redis.connection.mget(*cache_keys)
        objects: Dict[Any, Any] = OrderedDict()
        for id_value, data in zip(indices, res):
            objects[id_value] = cls.deserialize(data) if data is not None else None
        return list(objects.values())

    @classmethod
    async def _drop_one(cls, ctx: ContextData, cache_key: str) -> bool:
        return bool(await ctx.redis.connection.delete(cache_key))

    @classmethod
    async def _drop_many(cls, ctx: ContextData, indices: List[Any] = None) -> bool:
        if indices and len(indices) > 0:
            cache_keys = [cls.get_cache_key(id_value) for id_value in indices]
        else:
            model_cache_key = cls.get_model_cache_key()
            cache_keys = await ctx.redis.connection.keys(f'{model_cache_key}:*')

        return bool(await ctx.redis.connection.delete(*cache_keys)) if cache_keys else False

    @classmethod
    async def _exists(cls, ctx: ContextData, cache_key: str) -> bool:
        if await ctx.redis.connection.exists(cache_key):
            return await ctx.redis.connection.get(cache_key) != pickle.dumps(None)
        return False

    @classmethod
    async def _scan(cls, ctx: ContextData, cache_key: str = None,
                    match: str = None, count: int = None) -> Tuple[str, Optional[Dict[str, Any]]]:
        match = f'{cache_key}:{match}' if cache_key and match else match
        async for k in ctx.redis.connection.iscan(match=match, count=count):
            cache_key = k.decode('utf-8')
            yield cache_key.split(':')[1] if ':' in cache_key else cache_key, (await cls._load_one(ctx, cache_key))[1]


class HashRedisModelBase(RedisModelBase):
    @classmethod
    async def _cache_one(
        cls, ctx: ContextData, instance: Optional['RedisModelBase'] = None,
            id_value: Optional[Any] = None, data: Optional[Dict[str, Any]] = None,
            exp: Optional[timedelta] = None, **kwargs) -> bool:
        if instance and id_value is None:
            id_value = instance.get_id_value()

        model_cache_key, cache_id = cls.get_model_cache_key(), cls.get_cache_id(id_value)

        if data:
            data = pickle.dumps(data)
        else:
            data = instance.serialize() if instance else pickle.dumps(None)

        if cls.Meta.cache_ttl_enable:
            res = False
            tx = ctx.redis.connection.multi_exec()
            expiration = ctx.data.get(f'exp/{model_cache_key}', ctx.redis.engine.default_ttl)
            tx.hset(model_cache_key, cache_id, data)
            tx.expire(model_cache_key, int(expiration.total_seconds()))

            for r in await tx.execute():
                res |= r

            return bool(res)

        return bool(await ctx.redis.connection.hset(model_cache_key, cache_id, data))

    @classmethod
    async def _cache_many(
            cls, ctx: ContextData, objects: List['RedisModelBase'], exp: timedelta = None, **kwargs) -> bool:
        if not objects:
            return False
        model_cache_key = cls.get_model_cache_key()
        obj_dict = {cls.get_cache_id(instance.get_id_value()): instance.serialize() for instance in objects}

        if cls.Meta.cache_ttl_enable:
            res = False
            tx = ctx.redis.connection.multi_exec()
            expiration = ctx.data.get(f'exp/{model_cache_key}', ctx.redis.engine.default_ttl)
            tx.hmset_dict(model_cache_key, obj_dict)
            tx.expire(model_cache_key, int(expiration.total_seconds()))

            for r in await tx.execute():
                res |= r

            return bool(res)

        return bool(await ctx.redis.connection.hmset_dict(model_cache_key, obj_dict))

    @classmethod
    async def _load_one(
            cls, ctx: ContextData, cache_key: str, **kwargs) -> Tuple[bool, Optional[Dict[str, Any]]]:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':')
            if res := await ctx.redis.connection.hget(model_cache_key, cache_id):
                return True, cls.deserialize(res)
            return False, None
        else:
            if res := await ctx.redis.connection.get(cache_key):
                return True, cls.deserialize(res)
            return False, None

    @classmethod
    async def _load_many(
            cls, ctx: ContextData, indices: List[Any], **kwargs) -> List[Optional[Dict[str, Any]]]:
        if not indices:
            return []
        cache_keys = [cls.get_cache_id(id_value) for id_value in indices]
        res = await ctx.redis.connection.hmget(cls.get_model_cache_key(), *cache_keys)
        objects: Dict[Any, Any] = OrderedDict()
        for id_value, data in zip(indices, res):
            objects[id_value] = cls.deserialize(data) if data is not None else None
        return list(objects.values())

    @classmethod
    async def _drop_one(cls, ctx: ContextData, cache_key: str) -> bool:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':')
            return bool(await ctx.redis.connection.hdel(model_cache_key, cache_id))
        else:
            return bool(await ctx.redis.connection.delete(cache_key))

    @classmethod
    async def _drop_many(cls, ctx: ContextData, indices: List[Any] = None) -> bool:
        model_cache_key = cls.get_model_cache_key()

        if indices and len(indices) > 0:
            cache_keys = [cls.get_cache_id(id_value) for id_value in indices]
        else:
            cache_keys = await ctx.redis.connection.hkeys(model_cache_key)

        return bool(await ctx.redis.connection.hdel(model_cache_key, *cache_keys)) if cache_keys else False

    @classmethod
    async def _exists(cls, ctx: ContextData, cache_key: str) -> bool:
        if ':' in cache_key:
            model_cache_key, cache_id = cache_key.split(':')
            e = bool(await ctx.redis.connection.hexists(model_cache_key, cache_id))
        else:
            e = bool(await ctx.redis.connection.exists(cache_key))
        if e:
            return await ctx.redis.connection.get(cache_key) != pickle.dumps(None)
        return False

    @classmethod
    async def _scan(cls, ctx: ContextData, cache_key: str = None,
                    match: str = None, count: int = None) -> Tuple[str, Optional[Dict[str, Any]]]:
        async for k, data in ctx.redis.connection.ihscan(cache_key, match=match, count=count):
            yield k.decode('utf-8'), cls.deserialize(data)



try:
    import aioredis

    from dropland.storages.redis.engine import EngineConfig, RedisStorageBackend, RedisStorageEngine
    from dropland.storages.redis.connections import ConnectionManager
    from dropland.storages.redis.model import SimpleRedisModelBase as SimpleRedisModel, \
        HashRedisModelBase as HashRedisModel
    from dropland.storages.redis.settings import RedisSettings

    USE_REDIS = True

except ImportError:
    USE_REDIS = False



import contextlib
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Set, Tuple, TypeVar, Union

from dependency_injector.wiring import Provide, inject
from pydantic.main import BaseModel as PydanticModel
from sqlalchemy import Column, exists, func, literal_column, select, tuple_
from sqlalchemy.sql import ColumnCollection

from dropland.data.context import ContextData
from dropland.storages.base import StorageModel as BaseModel
from dropland.storages.sql import SqlStorageEngine, SqlStorageType
from dropland.storages.sql.connections import ConnectionManager

CreateSchemaType = TypeVar('CreateSchemaType', bound=PydanticModel)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=PydanticModel)


class SqlTableInfo(NamedTuple):
    db_type: SqlStorageType
    backend: str
    is_async: bool


class SqlModel(type):
    def __init__(cls, class_name, parents, attributes):
        super().__init__(class_name, parents, attributes)

    @inject
    def __new__(mcs, class_name, parents, attributes, manager: ConnectionManager = Provide['manager']):
        rv = super().__new__(mcs, class_name, parents, attributes)
        if isinstance(manager, ConnectionManager):
            rv._sql_manager = manager
        return rv


class SqlModelBase(BaseModel):
    _sql_engine: SqlStorageEngine = None
    _sql_manager: ConnectionManager = None

    # noinspection PyUnresolvedReferences
    @classmethod
    def _check_abstract(cls):
        if cls.__table__ is None:
            raise TypeError(f'Model {cls.__name__} is abstract, no table is defined!')

    def __eq__(self, other: 'SqlModelBase'):
        self._check_abstract()
        return type(self) == type(other) and self.get_id_value() == other.get_id_value()

    def __hash__(self):
        self._check_abstract()
        return self.get_id_value().__hash__()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.Meta.private_fields.update({'metadata', 'registry'})
        cls.Meta.non_serializable_fields.update({'metadata', 'registry'})

    def get_id_value(self) -> Any:
        return self._get_column_values(self._get_id_columns())

    @classmethod
    def get_engine(cls) -> SqlStorageEngine:
        return cls._sql_engine

    @classmethod
    def get_table_info(cls) -> SqlTableInfo:
        engine = cls.get_engine()
        return SqlTableInfo(db_type=engine.db_type, backend=engine.backend.name, is_async=engine.is_async)

    @classmethod
    @contextlib.contextmanager
    def _connection_context(cls, begin_tx: bool = False, autocommit: bool = False, ctx: ContextData = None):
        info, ctx = cls.get_table_info(), ctx if ctx else ContextData()
        assert not info.is_async, 'Use async_connection_context() function instead'
        if hasattr(ctx, 'sql'):
            yield ctx
        else:
            with cls._sql_manager.connection_context(info.db_type, info.backend, begin_tx, autocommit) as c:
                ctx.sql = c
                yield ctx
                del ctx.sql

    @classmethod
    @contextlib.asynccontextmanager
    async def _async_connection_context(cls, begin_tx: bool = False, autocommit: bool = False, ctx: ContextData = None):
        info, ctx = cls.get_table_info(), ctx if ctx else ContextData()
        assert info.is_async, 'Use connection_context() function instead'
        if hasattr(ctx, 'sql'):
            yield ctx
        else:
            async with cls._sql_manager.async_connection_context(info.db_type, info.backend, begin_tx, autocommit) as c:
                ctx.sql = c
                yield ctx
                del ctx.sql

    @classmethod
    def query_for_select(cls, **kwargs):
        raise NotImplementedError

    @classmethod
    def query_for_update(cls, **kwargs):
        raise NotImplementedError

    @classmethod
    def query_for_delete(cls, **kwargs):
        raise NotImplementedError

    # noinspection PyUnresolvedReferences
    @classmethod
    def get_columns(cls) -> ColumnCollection:
        return cls.__table__.columns

    #
    # Retrieve operations
    #

    @classmethod
    async def get(cls, id_value: Any, **kwargs) -> Optional['SqlModelBase']:
        query = cls.query_get(id_value, **kwargs)
        async with cls._async_connection_context() as ctx:
            if instance := await cls.perform_get(ctx, query, **kwargs):
                return await cls.construct_rela(ctx, instance, **kwargs)
        return None

    @classmethod
    async def list(cls,
                   filters: Optional[List[Any]] = None,
                   sorting: Optional[List[Any]] = None,
                   skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs) -> List['SqlModelBase']:
        query = cls.query_list(filters or [], sorting or [], skip, limit, params, **kwargs)
        async with cls._async_connection_context() as ctx:
            objects = await cls.perform_list(ctx, query, **kwargs)
            return await cls.construct_rela_list(ctx, objects, **kwargs)

    @classmethod
    async def get_any(cls, indices: List[Any], **kwargs) -> List[Optional['SqlModelBase']]:
        query = cls.query_any(indices, **kwargs)
        async with cls._async_connection_context() as ctx:
            objects = await cls.perform_any(ctx, query, **kwargs)
            objects = await cls.construct_rela_list(ctx, objects, **kwargs)
        if len(objects) == len(indices):
            return objects
        objects = {obj.get_id_value(): obj for obj in objects}
        return [objects[id_value] if id_value in objects else None for id_value in indices]

    @classmethod
    async def count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs) -> int:
        query = cls.query_count(filters, params, **kwargs)
        async with cls._async_connection_context() as ctx:
            return await cls.perform_count(ctx, query, **kwargs)

    @classmethod
    async def exists(cls, id_value: Any, **kwargs) -> bool:
        query = cls.query_exists(id_value, **kwargs)
        async with cls._async_connection_context() as ctx:
            return await cls.perform_exists(ctx, query, **kwargs)

    @classmethod
    async def exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs) -> bool:
        if not filters or not isinstance(filters, list):
            return False

        query = cls.query_exists_by(filters, params, **kwargs)
        async with cls._async_connection_context() as ctx:
            return await cls.perform_exists_by(ctx, query, **kwargs)

    #
    # Modification operations
    #

    @classmethod
    async def create(cls, data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) -> Optional['SqlModelBase']:
        if isinstance(data, dict):
            create_data = data
        else:
            create_data = data.dict(exclude_unset=True)

        if create_data:
            create_data = cls.prepare_for_create(create_data)

        async with cls._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            if instance := await cls.perform_create(ctx, create_data):
                return await cls.construct_rela(ctx, instance, **kwargs)
        return None

    @classmethod
    async def get_or_create(cls, id_value: Any,
                            data: Union[CreateSchemaType, Dict[str, Any]], **kwargs) \
            -> Tuple[Optional['SqlModelBase'], bool]:
        if instance := await cls.get(id_value, **kwargs):
            return instance, False
        else:
            return await cls.create(data, **kwargs), True

    @classmethod
    async def update_by_id(cls, id_value: Any, data: Union[UpdateSchemaType, Dict[str, Any]], **kwargs) \
            -> Optional['SqlModelBase']:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = cls.prepare_for_update(update_data)

        async with cls._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            if data := await cls.perform_update(ctx, update_data, id_value):
                if instance := await cls.construct(ctx, data, **kwargs):
                    return await cls.construct_rela(ctx, instance, **kwargs)
        return None

    async def update(self, data: Union[UpdateSchemaType, Dict[str, Any]], **kwargs) -> bool:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = self.prepare_for_update(update_data)

        async with self._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            if row := await self.perform_update(ctx, update_data, self.get_id_value()):
                await self.assign(row).reload_rela(ctx, set(update_data.keys()), **kwargs)
                return True

        return False

    @classmethod
    async def update_by(cls, filters: List[Any],
                        data: Union[UpdateSchemaType, Dict[str, Any]],
                        /, params: Dict[str, Any] = None, **kwargs) -> int:
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = cls.prepare_for_update(update_data)

        query = cls.query_update(filters, params, **kwargs)

        async with cls._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            return await cls.perform_update_by(ctx, update_data, query)

    @classmethod
    async def delete_by_id(cls, id_value: Any) -> bool:
        async with cls._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            return await cls.perform_delete(ctx, id_value)

    async def delete(self) -> bool:
        async with self._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            return await self.perform_delete(ctx, self.get_id_value())

    @classmethod
    async def delete_by(cls, filters: List[Any], /, params: Dict[str, Any] = None, **kwargs) -> int:
        query = cls.query_delete(filters, params, **kwargs)
        async with cls._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            return await cls.perform_delete_by(ctx, query)

    @classmethod
    def prepare_for_create(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        return data

    @classmethod
    def prepare_for_update(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        return data

    @classmethod
    def prepare_for_delete(cls, obj: 'SqlModelBase') -> 'SqlModelBase':
        return obj

    async def save(self, updated_fields: List[str] = None, **kwargs) -> bool:
        cols = self.get_columns()

        if updated_fields is not None and isinstance(updated_fields, Sequence):
            updated_fields = set(updated_fields)
            cols = [(col.name, col) for col in cols if self._get_field_by_column(col) in updated_fields]
            cols = ColumnCollection(columns=cols)
        else:
            updated_fields = set()

        col_values = self._get_column_values(cols, force_tuple=True)
        data = {k.name: v for k, v in zip(cols, col_values)}

        async with self._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            res = await self.perform_save(ctx, data, updated_fields, **kwargs)
            await self.reload_rela(ctx, updated_fields, **kwargs)

        return len(res.keys()) > 0 if res else False

    async def load(self, field_names: List[str] = None) -> bool:
        query = self.query_get(self.get_id_value())
        async with self._async_connection_context() as ctx:
            return await self.perform_load(ctx, query, field_names)

    @classmethod
    async def save_all(cls, objects: List['SqlModelBase'], **kwargs) -> bool:
        async with cls._async_connection_context(begin_tx=True, autocommit=True) as ctx:
            return await cls.perform_save_all(ctx, objects, **kwargs)

    #
    # Query operations
    #

    @classmethod
    def query_get(cls, id_value: Any, **kwargs):
        return cls._get_helper(id_value, **kwargs)

    @classmethod
    def query_list(cls, filters: List[Any], sorting: List[Any],
                   skip: int = 0, limit: int = 0, params: Dict[str, Any] = None, **kwargs):
        query = cls.query_for_select(**kwargs).offset(skip if skip >= 0 else 0)
        if limit > 0:
            query = query.limit(limit)
        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)
        for s in sorting:
            query = query.order_by(s)
        return query

    @classmethod
    def query_any(cls, indices: List[Any], **kwargs):
        return cls._get_any_helper(indices, **kwargs)

    # noinspection PyUnresolvedReferences
    @classmethod
    def query_count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs):
        if filters:
            query = cls.query_for_select()
            for f in filters:
                query = query.where(f)
            if params:
                query = query.params(**params)
        else:
            query = cls.__table__

        return select([func.count(literal_column('1'))]).select_from(query.alias())

    @classmethod
    def query_exists(cls, id_value: Any, **kwargs):
        return cls._get_helper(id_value, query=exists(), **kwargs).select()

    @classmethod
    def query_exists_by(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        query = exists(cls)

        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)

        return query.select()

    @classmethod
    def query_update(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        query = cls.query_for_update(**kwargs)

        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)

        return query

    @classmethod
    def query_delete(cls, filters: List[Any], params: Dict[str, Any] = None, **kwargs):
        query = cls.query_for_delete(**kwargs)

        for f in filters:
            query = query.where(f)
        if filters and params:
            query = query.params(**params)

        return query

    #
    # Perform operations
    #

    @classmethod
    async def perform_get(cls, ctx: ContextData, query, **kwargs) -> Optional['SqlModelBase']:
        raise NotImplementedError

    @classmethod
    async def perform_list(cls, ctx: ContextData, query, **kwargs) -> List['SqlModelBase']:
        raise NotImplementedError

    @classmethod
    async def perform_any(cls, ctx: ContextData, query, **kwargs) -> List[Optional['SqlModelBase']]:
        raise NotImplementedError

    @classmethod
    async def perform_count(cls, ctx: ContextData, query, **kwargs) -> int:
        raise NotImplementedError

    @classmethod
    async def perform_exists(cls, ctx: ContextData, query, **kwargs) -> bool:
        raise NotImplementedError

    @classmethod
    async def perform_exists_by(cls, ctx: ContextData, query, **kwargs) -> bool:
        raise NotImplementedError

    @classmethod
    async def perform_create(cls, ctx: ContextData, data: Dict[str, Any]) -> Optional['SqlModelBase']:
        raise NotImplementedError

    @classmethod
    async def perform_update(
            cls, ctx: ContextData, data: Dict[str, Any], id_value: Any) -> Optional[Dict[str, Any]]:
        raise NotImplementedError

    @classmethod
    async def perform_update_by(cls, ctx: ContextData, data: Dict[str, Any], query) -> int:
        raise NotImplementedError

    @classmethod
    async def perform_delete(cls, ctx: ContextData, id_value: Any) -> bool:
        raise NotImplementedError

    @classmethod
    async def perform_delete_by(cls, ctx: ContextData, query) -> int:
        raise NotImplementedError

    async def perform_save(
        self, ctx: ContextData, data: Dict[str, Any], updated_fields: Set[str], **kwargs) \
            -> Optional[Dict[str, Any]]:
        raise NotImplementedError

    async def perform_load(self, ctx: ContextData, query, field_names: List[str] = None) -> bool:
        raise NotImplementedError

    @classmethod
    async def perform_save_all(cls, ctx: ContextData, objects: List['SqlModelBase'], **kwargs) -> bool:
        raise NotImplementedError

    #
    # Private
    #

    # noinspection PyUnresolvedReferences
    @classmethod
    def _get_id_columns(cls) -> ColumnCollection:
        return cls.__table__.primary_key.columns

    # noinspection PyTypeChecker
    def _get_column_values(
            self, columns: ColumnCollection, force_tuple: bool = False) -> Union[Any, Tuple[Any]]:
        rv = []
        for c in columns:
            rv.append(getattr(self, self._get_field_by_column(c)))
        return rv[0] if len(rv) == 1 and not force_tuple else tuple(rv)

    @classmethod
    def _get_field_by_column(cls, c: Column) -> str:
        raise NotImplementedError

    # noinspection PyUnresolvedReferences
    @classmethod
    def _get_fk_columns(cls, model) -> ColumnCollection:
        for fk in cls.__table__.foreign_keys:
            if model.__table__ == fk.constraint.referred_table:
                return fk.constraint.columns
        return ColumnCollection()

    @classmethod
    def _get_helper(cls, ident: Any, query=None, **kwargs):
        cls._check_abstract()
        if not isinstance(ident, (list, tuple, dict)):
            ident_ = [ident]
        else:
            ident_ = ident
        columns = cls._get_id_columns()
        if len(ident_) != len(columns):
            raise ValueError(
                f'Incorrect number of values as primary key: expected {len(columns)}, got {len(ident_)}.')

        clause = query if query is not None else cls.query_for_select(**kwargs)
        for i, c in enumerate(columns):
            try:
                val = ident_[i]
            except KeyError:
                val = ident_[cls._get_field_by_column(c)]
            clause = clause.where(c == val)
        return clause

    @classmethod
    def _get_any_helper(cls, indices: List[Any], query=None, **kwargs):
        cls._check_abstract()
        columns = cls._get_id_columns()
        clause = query if query is not None else cls.query_for_select(**kwargs)
        vals_clause = []

        for ident in indices:
            if not isinstance(ident, (list, tuple, dict)):
                ident_ = [ident]
            else:
                ident_ = ident

            if len(ident_) != len(columns):
                raise ValueError(
                    f'Incorrect number of values as primary key: expected {len(columns)}, got {len(ident_)}.')

            vals = []
            for i, c in enumerate(columns):
                try:
                    val = ident_[i]
                except KeyError:
                    val = ident_[cls._get_field_by_column(c)]
                vals.append(val)

            if len(vals) == 1:
                vals_clause.append(vals[0])
            elif len(vals) > 1:
                vals_clause.append((*vals,))

        if len(columns) == 1:
            clause = clause.where(columns[columns.keys()[0]].in_(vals_clause))
        elif len(columns) > 1:
            clause = clause.where(tuple_(*columns).in_(vals_clause))
        return clause

    @classmethod
    async def _get_from_orm_cache(cls, ctx: ContextData, indices: List[Any]) -> List[Optional['SqlModelBase']]:
        return [None for _ in indices]

    @classmethod
    async def _attach_to_session(cls, ctx: ContextData, objects: List['SqlModelBase']):
        pass



import contextlib
import enum
from dataclasses import dataclass
from datetime import timedelta
from typing import List, Optional

from dropland.storages.base import StorageBackend, StorageEngine

# Recommended naming convention used by Alembic, as various different database
# providers will autogenerate vastly different names making migrations more
# difficult. See: http://alembic.zzzcomputing.com/en/latest/naming.html
NAMING_CONVENTION = {
    'ix': 'ix_%(column_0_label)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'pk': 'pk_%(table_name)s'
}


class SqlStorageType(str, enum.Enum):
    SQLITE = 'sqlite'
    POSTGRES = 'postgresql'
    MYSQL = 'mysql'


@dataclass
class EngineConfig:
    url: str
    echo: bool = False
    pool_min_size: int = 1
    pool_max_size: int = 8
    pool_expire_seconds: int = 60
    pool_timeout_seconds: int = 15


class SqlStorageBackend(StorageBackend):
    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def db_supports(self) -> List[SqlStorageType]:
        raise NotImplementedError

    @property
    def connection_class(self) -> Optional[type]:
        raise NotImplementedError

    @property
    def async_connection_class(self) -> Optional[type]:
        raise NotImplementedError

    def create_engine(self, db_type: SqlStorageType, config: EngineConfig, use_async: bool) \
            -> Optional['SqlStorageEngine']:
        raise NotImplementedError


class SqlStorageEngine(StorageEngine):
    def __init__(self, backend: SqlStorageBackend, raw_engine, db_type: SqlStorageType, timeout: timedelta):
        super().__init__(backend)
        self._engine = raw_engine
        self._db_type = db_type
        self._timeout = timeout

    @property
    def raw_engine(self):
        return self._engine

    @property
    def db_type(self):
        return self._db_type

    @property
    def timeout(self) -> timedelta:
        return self._timeout

    # noinspection PyPep8Naming
    @property
    def Model(self) -> 'Model':
        raise NotImplementedError

    @property
    def metadata(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def transaction_context(self, connection, autocommit: bool = True):
        raise NotImplementedError

    @contextlib.asynccontextmanager
    async def async_transaction_context(self, connection, autocommit: bool = True):
        raise NotImplementedError



try:
    from dropland.storages.sql.base import EngineConfig, SqlStorageBackend, SqlStorageEngine, SqlStorageType
    from dropland.storages.sql.connections import ConnectionManager
    from dropland.storages.sql.model import SqlModel, SqlModelBase
    from dropland.storages.sql.settings import SqliteSettings, PgSettings, MySqlSettings

    USE_SQL = True

except ImportError:
    USE_SQL = False



import asyncio
import threading
from concurrent.futures import Executor
from functools import partial
from typing import Callable, Optional, Union

from apscheduler.schedulers.asyncio import AsyncIOScheduler as BaseAsyncScheduler
from apscheduler.util import get_callable_name, iscoroutinefunction_partial, obj_to_ref, ref_to_obj, undefined

from dropland.storages.containers import use_engines
from dropland.tasks.remote import RemoteScheduler


class Scheduler(BaseAsyncScheduler):
    def __init__(self, remote: Optional[RemoteScheduler] = None,
                 executor: Optional[Executor] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._remote = remote
        self._executor = executor
        self._lock = threading.Lock()
        self._counter = 0
        self._thread = None

        try:
            self._eventloop = asyncio.get_running_loop()
        except RuntimeError:
            self._eventloop = asyncio.new_event_loop()

        self._eventloop.set_default_executor(self._executor)

    def _run(self):
        try:
            self._eventloop.run_forever()
        finally:
            self._eventloop.run_until_complete(self._eventloop.shutdown_asyncgens())
            self._eventloop.run_until_complete(self._eventloop.shutdown_default_executor())
            self._eventloop.close()

    def _start(self, paused: bool):
        if not self._eventloop.is_running() and not self._thread:
            self._thread = threading.Thread(target=self._run, name='TaskWorkerAsync')

        if self._thread and not self._thread.is_alive():
            self._thread.start()

        if self._remote:
            self._remote.start(paused)
        else:
            super().start(paused)

    def _shutdown(self, wait: bool):
        if self._remote:
            self._remote.shutdown(wait)
        else:
            super().shutdown(wait)

        if self._thread:
            self._eventloop.call_soon_threadsafe(self._eventloop.stop)
            self._thread.join()
            self._thread = None

    def start(self, paused: bool = False):
        with self._lock:
            if 0 == self._counter:
                self._start(paused)
            self._counter += 1

    def shutdown(self, wait: bool = True):
        with self._lock:
            if 1 == self._counter:
                self._shutdown(wait)
            self._counter = max(self._counter - 1, 0)

    def pause(self):
        if self._remote:
            self._remote.pause()
        else:
            super().pause()

    def resume(self):
        if self._remote:
            self._remote.resume()
        else:
            super().resume()

    def add_job(self, func: Union[str, Callable], trigger=None, args=None, kwargs=None, id=None, name=None,
                misfire_grace_time=undefined, coalesce=undefined, max_instances=undefined,
                next_run_time=undefined, jobstore='default', executor='default',
                replace_existing=False, **trigger_args):
        if self._remote:
            method = self._remote.add_job
            target = func
            fname = name or get_callable_name(func)
        else:
            func = func if isinstance(func, Callable) else ref_to_obj(func)

            if iscoroutinefunction_partial(func):
                target = Scheduler.task_sync_main if 'process' == executor else Scheduler.task_async_main
                args = (obj_to_ref(func), *(args or tuple()))
                fname = name or get_callable_name(func)
            else:
                target = func
                fname = name or get_callable_name(func)

            method = super().add_job

        return method(
            target, trigger, args, kwargs, id, fname, misfire_grace_time, coalesce,
            max_instances, next_run_time, jobstore, executor, replace_existing, **trigger_args)

    def scheduled_job(self, trigger, args=None, kwargs=None, id=None, name=None,
                      misfire_grace_time=undefined, coalesce=undefined, max_instances=undefined,
                      next_run_time=undefined, jobstore='default', executor='default', **trigger_args):
        return super().scheduled_job(
            trigger, args, kwargs, id, name, misfire_grace_time, coalesce,
            max_instances, next_run_time, jobstore, executor, **trigger_args)

    def modify_job(self, job_id, jobstore=None, **changes) -> bool:
        if _ := super().get_job(job_id, jobstore):
            super().modify_job(job_id, jobstore, **changes)
            return True
        if self._remote:
            return self._remote.modify_job(job_id, jobstore, **changes)
        return False

    def reschedule_job(self, job_id, jobstore=None, trigger=None, **trigger_args) -> bool:
        if super().reschedule_job(job_id, jobstore, trigger, **trigger_args):
            return True
        if self._remote:
            return self._remote.reschedule_job(job_id, jobstore, trigger, **trigger_args)
        return False

    def pause_job(self, job_id, jobstore=None) -> bool:
        if _ := super().get_job(job_id, jobstore):
            super().pause_job(job_id, jobstore)
            return True
        if self._remote:
            return self._remote.pause_job(job_id, jobstore)
        return False

    def resume_job(self, job_id, jobstore=None) -> bool:
        if _ := super().get_job(job_id, jobstore):
            super().resume_job(job_id, jobstore)
            return True
        if self._remote:
            return self._remote.resume_job(job_id, jobstore)
        return False

    def remove_job(self, job_id, jobstore=None) -> bool:
        if _ := super().get_job(job_id, jobstore):
            super().remove_job(job_id, jobstore)
            return True
        if self._remote:
            return self._remote.remove_job(job_id, jobstore)
        return False

    def remove_all_jobs(self, jobstore=None):
        if self._remote:
            self._remote.remove_all_jobs(jobstore)
        else:
            super().remove_all_jobs(jobstore)

    def get_job(self, job_id, jobstore=None):
        if job := super().get_job(job_id, jobstore):
            return job
        if self._remote:
            return self._remote.get_job(job_id, jobstore)
        return None

    def get_jobs(self, jobstore=None, pending=None):
        if jobs := super().get_jobs(jobstore, pending):
            return jobs
        if self._remote:
            return self._remote.get_jobs(jobstore, pending)
        return []

    def lookup_executor(self, alias: str):
        return self._executors.get(alias)

    @staticmethod
    async def task_async_main(func: Union[str, Callable], *args, **kwargs):
        func = func if isinstance(func, Callable) else ref_to_obj(func)
        target = use_engines()(func)

        if iscoroutinefunction_partial(func):
            await target(*args, **kwargs)
        else:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, partial(target, *args, **kwargs))

    @staticmethod
    def task_sync_main(func: Union[str, Callable], *args, **kwargs):
        asyncio.run(Scheduler.task_async_main(func, *args, **kwargs))



from typing import Dict, Any


class ContextData(object):
    def __init__(self, props: Dict[str, Any] = None, **kwargs):
        if props:
            for name, value in props.items():
                super().__setattr__(name, self._wrap(value))
        super().__setattr__('data', kwargs)

    def _wrap(self, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([self._wrap(v) for v in value])
        else:
            return ContextData(value) if isinstance(value, dict) else value

    def __setattr__(self, key, value):
        super().__setattr__(key, self._wrap(value))



import contextlib
from typing import Any, Dict, List, Optional, Set, Type

from dropland.core.metaclass import classmaker
from dropland.data.context import ContextData
from dropland.storages import CacheModel
from dropland.storages.sql import SqlModelBase


class CachedModelImpl:
    @classmethod
    def has_cache(cls):
        return True

    # noinspection PyUnresolvedReferences
    @classmethod
    def get_engine(cls):
        return cls._sql_class.get_engine()

    # noinspection PyUnresolvedReferences,PyProtectedMember
    @classmethod
    @contextlib.contextmanager
    def _connection_context(cls, begin_tx: bool = False, autocommit: bool = False, ctx: ContextData = None):
        with cls._sql_class._connection_context(begin_tx, autocommit, ctx) as ctx:
            yield ctx

    # noinspection PyUnresolvedReferences,PyProtectedMember
    @classmethod
    @contextlib.asynccontextmanager
    async def _async_connection_context(cls, begin_tx: bool = False, autocommit: bool = False, ctx: ContextData = None):
        async with cls._sql_class._async_connection_context(begin_tx, autocommit, ctx) as ctx:
            async with cls._cache_class._async_connection_context(ctx) as ctx:
                yield ctx

    #
    # Query operations
    #

    # noinspection PyUnresolvedReferences
    @classmethod
    def query_get(cls, id_value: Any, **kwargs):
        return id_value, super().query_get(id_value, **kwargs)

    # noinspection PyUnresolvedReferences
    @classmethod
    def query_any(cls, indices: List[Any], **kwargs):
        return indices, super().query_any(indices, **kwargs)

    # noinspection PyUnresolvedReferences
    @classmethod
    def query_count(cls, filters: Optional[List[Any]] = None, params: Dict[str, Any] = None, **kwargs):
        if not filters or not isinstance(filters, list):
            cache_key = f'{cls.get_model_cache_key()}.count'
        else:
            cache_key = None

        return cache_key, super().query_count(filters, params, **kwargs)

    # noinspection PyUnresolvedReferences
    @classmethod
    def query_exists(cls, id_value: Any, **kwargs):
        return cls.get_cache_key(id_value), super().query_exists(id_value, **kwargs)

    #
    # Perform operations
    #

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_get(cls, ctx: ContextData, query, **kwargs) -> Optional['CachedModelImpl']:
        id_value, query = query[0], query[1]
        exists, data = await cls._load_one(ctx, cls.get_cache_key(id_value), **kwargs)
        if exists:
            if data is not None:
                if orm_cached := (await cls._get_from_orm_cache(ctx, [id_value]))[0]:
                    return orm_cached
            return await cls.construct(ctx, data, **kwargs)

        instance = await super().perform_get(ctx, query, **kwargs)
        await cls._cache_one(ctx, instance, id_value)
        return instance

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_list(cls, ctx: ContextData, query, **kwargs) -> List['CachedModelImpl']:
        objects = await super().perform_list(ctx, query, **kwargs)
        await cls._cache_many(ctx, objects, **kwargs)
        await cls._drop_one(ctx, f'{cls.get_model_cache_key()}.count')
        return objects

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_any(cls, ctx: ContextData, query, **kwargs) -> List[Optional['CachedModelImpl']]:
        indices, query = query[0], query[1]
        cached_indices, non_cached_indices = list(), list()

        cached_objects = await cls._load_many(ctx, indices, **kwargs)

        for id_value, instance in zip(indices, cached_objects):
            if instance is None:
                non_cached_indices.append(id_value)
            else:
                cached_indices.append(id_value)

        orm_cached_objects = {
            id_value: obj for id_value, obj in
            zip(cached_indices, await cls._get_from_orm_cache(ctx, cached_indices))
        }

        for id_value, (i, instance) in zip(indices, enumerate(cached_objects)):
            if id_value in orm_cached_objects:
                if obj := orm_cached_objects[id_value]:
                    cached_objects[i] = obj

        if not non_cached_indices:
            return await cls.construct_list(ctx, cached_objects, **kwargs)
        elif cached_objects:
            query = super().query_any(non_cached_indices, **kwargs)

        non_cached_objects = await super().perform_any(ctx, query, **kwargs)
        await cls._cache_many(ctx, non_cached_objects, **kwargs)

        cached_objects = await cls.construct_list(ctx, cached_objects, **kwargs)
        cached_objects = {id_value: obj for id_value, obj in zip(indices, cached_objects) if obj is not None}
        non_cached_objects = {obj.get_id_value(): obj for obj in non_cached_objects}
        result = []

        for id_value in indices:
            if id_value in non_cached_objects:
                result.append(non_cached_objects[id_value])
            elif id_value in cached_objects:
                result.append(cached_objects[id_value])

        return result

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_count(cls, ctx: ContextData, query, **kwargs) -> int:
        cache_key, query = query[0], query[1]

        if cache_key:
            res = await ctx.redis.connection.get(cache_key)
            if res is not None:
                return int(res)

        res = await super().perform_count(ctx, query, **kwargs)

        if res and cache_key:
            cache_kwargs = dict()

            if cls.Meta.cache_ttl_enable:
                total_seconds = ctx.redis.engine.default_ttl.total_seconds()
                cache_kwargs['expire'] = int(total_seconds) if total_seconds > 1 else 60

            await ctx.redis.connection.set(cache_key, res, **cache_kwargs)

        return res

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_exists(cls, ctx: ContextData, query, **kwargs) -> bool:
        cache_key, query = query[0], query[1]
        if await cls._exists(ctx, cache_key):
            return True
        return await super().perform_exists(ctx, query, **kwargs)

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_exists_by(cls, ctx: ContextData, query, **kwargs) -> bool:
        return await super().perform_exists_by(ctx, query, **kwargs)

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_create(cls, ctx: ContextData, data: Dict[str, Any]) -> Optional['CachedModelImpl']:
        instance = await super().perform_create(ctx, data)
        await cls._cache_one(ctx, instance)
        await cls._drop_one(ctx, f'{cls.get_model_cache_key()}.count')
        return instance

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_update(
            cls, ctx: ContextData, data: Dict[str, Any], id_value: Any) -> Optional[Dict[str, Any]]:
        data = await super().perform_update(ctx, data, id_value)
        await cls._drop_one(ctx, cls.get_cache_key(id_value))
        return data

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_update_by(cls, ctx: ContextData, data: Dict[str, Any], query) -> int:
        res = await super().perform_update_by(ctx, data, query)
        await cls._drop_many(ctx)
        return res

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_delete(cls, ctx: ContextData, id_value: Any) -> bool:
        if res := await super().perform_delete(ctx, id_value):
            await cls._drop_one(ctx, cls.get_cache_key(id_value))
            await cls._drop_one(ctx, f'{cls.get_model_cache_key()}.count')
        return res

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_delete_by(cls, ctx: ContextData, query) -> int:
        res = await super().perform_delete_by(ctx, query)
        await cls._drop_many(ctx)
        return res

    # noinspection PyUnresolvedReferences
    async def perform_save(
        self, ctx: ContextData, data: Dict[str, Any], updated_fields: Set[str], **kwargs) \
            -> Optional[Dict[str, Any]]:
        if data := await super().perform_save(ctx, data, updated_fields, **kwargs):
            await self._cache_one(ctx, self, data=data)
        return data

    # noinspection PyUnresolvedReferences
    async def perform_load(self, ctx: ContextData, query, field_names: List[str] = None) -> bool:
        id_value, query = query[0], query[1]
        exists, data = await self._load_one(ctx, self.get_cache_key(id_value))
        if exists and isinstance(data, dict):
            await self.assign(data).reload_rela(ctx, field_names)
            return True

        if res := await super().perform_load(ctx, query, field_names):
            await self._cache_one(ctx, self)
        return res

    # noinspection PyUnresolvedReferences
    @classmethod
    async def perform_save_all(cls, ctx: ContextData, objects: List['CachedModelImpl'], **kwargs) -> bool:
        if res := await super().perform_save_all(ctx, objects, **kwargs):
            await cls._cache_many(ctx, objects, **kwargs)
            await cls._drop_one(ctx, f'{cls.get_model_cache_key()}.count')
        return res


# noinspection PyPep8Naming
def CachedModel(cache_name: str, sql_class: Type[SqlModelBase], cache_class: Type[CacheModel]):
    # noinspection PyShadowingNames,PyUnresolvedReferences
    class CachedModel(sql_class.__metaclass__, cache_class.__metaclass__):
        # noinspection PyUnresolvedReferences
        @property
        def _model_cache_key(self):
            return f'{cache_name}.models.{self.__tablename__}'

    # noinspection PyAbstractClass
    class CachedModelClass(CachedModelImpl, sql_class, cache_class, metaclass=CachedModel):
        __abstract__ = True
        __metaclass__ = classmaker()
        _sql_class = sql_class
        _cache_class = cache_class

    return CachedModelClass



import pytest

from tests.sql_models_data import USE_GINO

pytestmark = pytest.mark.skipif(not USE_GINO, reason='For Gino only')

if USE_GINO:
    from dropland.storages.containers import default_sql_storage
    from dropland.storages.sql import SqlStorageType


@pytest.mark.asyncio
async def test_init_context():
    mng = default_sql_storage.manager()

    @mng.use_engines(['gino'])
    async def async_context_func():
        assert not mng.get_connection(SqlStorageType.POSTGRES, 'gino', False)
        assert mng.get_connection(SqlStorageType.POSTGRES, 'gino', True)

        assert 0 == len(mng.get_connections_for_type(SqlStorageType.POSTGRES, False))

        for c in mng.get_connections_for_type(SqlStorageType.POSTGRES, True):
            assert c.engine
            assert c.connection
            assert c.in_transaction

        assert 0 == len(mng.get_connections_for_backend('gino', False))

        for c in mng.get_connections_for_backend('gino', True):
            assert c.engine
            assert c.connection
            assert c.in_transaction

        return mng

    await async_context_func()


@pytest.mark.asyncio
async def test_connections(pg_async_engine, mysql_async_engine):
    mng = default_sql_storage.manager()

    @mng.use_engines(['gino'])
    async def async_context_func():
        created, sql_data = mng.get_or_create_connection(SqlStorageType.POSTGRES, 'gino', False)
        assert not created
        assert not sql_data

        created, sql_data = mng.get_or_create_connection(SqlStorageType.POSTGRES, 'gino', True)
        assert not created
        assert sql_data

        assert sql_data.engine == pg_async_engine

        created, sql_data = mng.get_or_create_connection(SqlStorageType.MYSQL, 'gino', True)
        assert not created
        assert sql_data

        assert sql_data.engine == mysql_async_engine

    await async_context_func()



import pytest

from dropland.storages.redis import USE_REDIS

pytestmark = pytest.mark.skipif(not USE_REDIS, reason='For Redis only')


@pytest.mark.asyncio
async def test_init_context(test_redis_storage, redis_engine):
    mng = test_redis_storage.manager()

    @mng.use_engines(['dropland'])
    async def context_func():
        assert not mng.get_connection('_')

        redis_data = mng.get_connection('dropland')
        assert redis_data
        assert redis_data.engine
        assert redis_data.connection

    await context_func()


@pytest.mark.asyncio
async def test_connections(test_redis_storage, redis_engine):
    mng = test_redis_storage.manager()

    @mng.use_engines(['dropland'])
    async def context_func():
        created, redis_data = mng.get_or_create_connection('dropland')
        assert not created
        assert redis_data

        assert redis_data.engine == redis_engine
        assert redis_data.connection

    await context_func()



from importlib.metadata import entry_points
from typing import Optional, TypeVar, Dict, Generic

BackendType = TypeVar('BackendType')


class BackendLoader(Generic[BackendType]):
    def __init__(self, ep_name: str, ep_function: str):
        self._load = False
        self._ep_name = ep_name
        self._ep_function = ep_function
        self._backends: Dict[str, BackendType] = dict()

    def load_backends(self):
        if self._load:
            return

        for bb in entry_points().get(self._ep_name, []):
            path = f'{self._ep_name}.backends.{bb.name}'

            # noinspection PyBroadException
            try:
                module = __import__(path, fromlist=f'{self._ep_name}.backends')
                if hasattr(module, self._ep_function):
                    backend = getattr(module, self._ep_function)()
                    self._backends[backend.name] = backend
                    self.on_load(backend)

            except Exception as e:
                continue

        self._load = True

    def get_backend(self, name: str) -> Optional[BackendType]:
        return self._backends.get(name)

    def on_load(self, backend: BackendType):
        pass



import pytest

from dropland.storages.redis import USE_REDIS

pytestmark = pytest.mark.skipif(not USE_REDIS, reason='For Redis only')

if USE_REDIS:
    from dropland.storages.containers import RedisStorage, SingleRedisStorage, MultipleRedisStorage, \
        default_redis_storage
    from dropland.storages.redis import EngineConfig, RedisStorageEngine
    from tests import REDIS_URI


@pytest.mark.asyncio
async def test_create_engine():
    engine_factory = default_redis_storage.engine_factory()

    assert not engine_factory.get_engine('')

    config = EngineConfig(url=REDIS_URI)
    assert engine_factory.create_engine('dropland', config)

    engine = engine_factory.get_engine('dropland')
    assert engine
    assert engine.backend is engine_factory
    assert engine.is_async

    assert engine_factory.get_engine_names() == ['dropland']


@pytest.mark.asyncio
async def test_create_connection():
    engine_factory = default_redis_storage.engine_factory()

    config = EngineConfig(url=REDIS_URI)
    engine = engine_factory.create_engine('dropland', config)

    assert engine
    await engine.async_start()

    async with engine.new_connection() as conn:
        assert await conn.set('kkk', 'vvv') is True
        assert b'vvv' == await conn.get('kkk')
        assert await conn.delete('kkk')

    await engine.async_stop()


@pytest.mark.asyncio
async def test_storage_container():
    config = EngineConfig(url=REDIS_URI)

    cont = RedisStorage()

    eng = cont.create_engine('dropland', config)
    assert isinstance(eng, RedisStorageEngine)
    assert eng.name == 'dropland'
    cont.unwire()

    cont = SingleRedisStorage()
    cont.config.from_dict({
        'name': 'dropland1',
        'engine_config': config
    })
    eng1 = cont.create_engine()
    eng2 = cont.create_engine()
    assert eng1 is eng2
    cont.unwire()

    cont = MultipleRedisStorage()
    cont.config.from_dict({
        'one': {
            'name': 'dropland1',
            'engine_config': config
        },
        'two': {
            'name': 'dropland2',
            'engine_config': config
        },
    })

    eng1 = cont.create_engine('one')
    eng2 = cont.create_engine('two')
    assert eng1 is not eng2
    assert eng1.is_async is True
    assert eng2.is_async is True
    assert eng1.name == 'dropland1'
    assert eng2.name == 'dropland2'

    cont.unwire()



from dropland.storages.sql import USE_SQL

if USE_SQL:
    pass

USE_SQLA = True
USE_GINO = True

try:
    # noinspection PyUnresolvedReferences
    from dropland.storages.sql.backends.sqla import SqlaModel

except ImportError:
    USE_SQLA = False

try:
    # noinspection PyUnresolvedReferences
    from dropland.storages.sql.backends.gino import GinoModel

except ImportError:
    USE_GINO = False


if USE_SQL:
    from tests.conftest import sql_storage

    sqla_sqlite = sql_storage.create_engine('asqlite')
    sqla_pg = sql_storage.create_engine('apg')
    sqla_mysql = sql_storage.create_engine('ams')
    gino_pg = sql_storage.create_engine('apg-gino')
    gino_mysql = sql_storage.create_engine('ams-gino')



try:
    from apscheduler.schedulers.base import BaseScheduler

    from dropland.tasks.engine import EngineConfig, TaskManagerBackend
    from dropland.tasks.local import Scheduler
    from dropland.tasks.settings import SchedulerSettings

    USE_SCHEDULER = True

except ImportError:
    USE_SCHEDULER = False
